"""
Word2Vec implementation supporting both Skip-gram and CBOW architectures
based on "Efficient Estimation of Word Representations in Vector Space"
https://arxiv.org/abs/1301.3781
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import Dataset, DataLoader
import numpy as np
from collections import Counter, deque
from typing import List, Dict, Tuple, Optional, Union, Iterator, Callable
import logging
import random
from tqdm import tqdm

logging.basicConfig(
    format='%(asctime)s : %(levelname)s : %(message)s',
    level=logging.INFO
)

class Word2VecDataset(Dataset):
    """Dataset for generating training samples for Word2Vec"""
    
    def __init__(
        self,
        texts: List[List[str]],
        word_to_idx: Dict[str, int],
        window_size: int = 5,
        mode: str = 'skipgram',
        ns_exponent: float = 0.75,
        negative_samples: int = 5,
        min_count: int = 5,
        subsampling_threshold: float = 1e-5
    ):
        """
        Initialize Word2Vec dataset
        
        Args:
            texts: List of tokenized sentences (list of list of tokens)
            word_to_idx: Mapping from words to indices
            window_size: Context window size
            mode: Training mode ('skipgram' or 'cbow')
            ns_exponent: Exponent for negative sampling distribution
            negative_samples: Number of negative samples per positive sample
            min_count: Minimum frequency for a word to be included
            subsampling_threshold: Threshold for subsampling frequent words
        """
        self.texts = texts
        self.word_to_idx = word_to_idx
        self.idx_to_word = {idx: word for word, idx in self.word_to_idx.items()}
        self.vocab_size = len(word_to_idx)
        self.window_size = window_size
        self.mode = mode.lower()
        self.negative_samples = negative_samples
        
        # Prepare data
        self.data = []
        self._prepare_training_data()
        
        # Prepare negative sampling distribution
        word_counts = Counter()
        for text in texts:
            word_counts.update([w for w in text if w in self.word_to_idx])
            
        self.word_counts = {self.word_to_idx[word]: count for word, count in word_counts.items() 
                            if count >= min_count}
        
        # Prepare subsampling
        self.total_words = sum(self.word_counts.values())
        self.discard_probs = {}
        for word_idx, count in self.word_counts.items():
            frequency = count / self.total_words
            # Formula from original Word2Vec paper
            self.discard_probs[word_idx] = 1.0 - np.sqrt(subsampling_threshold / frequency)
        
        # Prepare negative sampling distribution
        ns_counts = {idx: count ** ns_exponent for idx, count in self.word_counts.items()}
        total = sum(ns_counts.values())
        self.ns_prob = {idx: count / total for idx, count in ns_counts.items()}
        self.ns_dist = list(self.ns_prob.keys())
        self.ns_weights = list(self.ns_prob.values())
    
    def _prepare_training_data(self):
        """Process raw texts into training pairs"""
        for text in self.texts:
            indices = [self.word_to_idx[word] for word in text if word in self.word_to_idx]
            if len(indices) < 2:
                continue
                
            # Apply subsampling
            if hasattr(self, 'discard_probs'):
                indices = [idx for idx in indices if 
                          idx not in self.discard_probs or 
                          random.random() > self.discard_probs[idx]]
            
            for i, idx in enumerate(indices):
                # Define context window boundaries
                start = max(0, i - self.window_size)
                end = min(len(indices), i + self.window_size + 1)
                
                # Get context words
                context_indices = [indices[j] for j in range(start, end) if j != i]
                
                if not context_indices:
                    continue
                    
                if self.mode == 'skipgram':
                    # For skip-gram, each context word becomes a sample
                    for context_idx in context_indices:
                        self.data.append((idx, context_idx))
                else:  # CBOW
                    # For CBOW, all context words together form one sample
                    self.data.append((context_indices, idx))
    
    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, idx):
        sample = self.data[idx]
        
        if self.mode == 'skipgram':
            target_idx, context_idx = sample
            # Generate negative samples
            neg_samples = self._get_negative_samples(context_idx)
            return target_idx, context_idx, neg_samples
        else:  # CBOW
            context_indices, target_idx = sample
            # For CBOW we'll average the context vectors
            context_indices = torch.tensor(context_indices, dtype=torch.long)
            # Generate negative samples
            neg_samples = self._get_negative_samples(target_idx)
            return context_indices, target_idx, neg_samples
    
    def _get_negative_samples(self, positive_idx):
        """Generate negative samples, making sure to exclude the positive sample"""
        samples = []
        while len(samples) < self.negative_samples:
            sample = np.random.choice(self.ns_dist, p=self.ns_weights)
            if sample != positive_idx and sample not in samples:
                samples.append(sample)
        return torch.tensor(samples, dtype=torch.long)


class Word2Vec(nn.Module):
    """Word2Vec model supporting both Skip-gram and CBOW architectures"""
    
    def __init__(
        self,
        vocab_size: int,
        embedding_dim: int = 300,
        mode: str = 'skipgram',
        padding_idx: Optional[int] = None,
        sparse: bool = False
    ):
        """
        Initialize Word2Vec model
        
        Args:
            vocab_size: Size of vocabulary
            embedding_dim: Dimension of word embeddings
            mode: Model architecture ('skipgram' or 'cbow')
            padding_idx: If specified, padded embeddings will remain zeroed
            sparse: If True, gradients for embeddings will be sparse
        """
        super().__init__()
        self.vocab_size = vocab_size
        self.embedding_dim = embedding_dim
        self.mode = mode.lower()
        
        # Input embeddings (target words in skip-gram, context in CBOW)
        self.in_embeddings = nn.Embedding(
            vocab_size, embedding_dim, padding_idx=padding_idx, sparse=sparse
        )
        
        # Output embeddings (context words in skip-gram, target in CBOW)
        self.out_embeddings = nn.Embedding(
            vocab_size, embedding_dim, padding_idx=padding_idx, sparse=sparse
        )
        
        # Initialize embeddings
        self._init_weights()
    
    def _init_weights(self):
        """Initialize embeddings using uniform distribution as in the paper"""
        initrange = 0.5 / self.embedding_dim
        self.in_embeddings.weight.data.uniform_(-initrange, initrange)
        self.out_embeddings.weight.data.uniform_(-initrange, initrange)
    
    def forward_skipgram(self, target_idx, context_idx, neg_samples):
        """Forward pass for Skip-gram architecture"""
        # Get embeddings for target word
        target_emb = self.in_embeddings(target_idx)  # [batch_size, embed_dim]
        
        # Get embeddings for positive context word
        context_emb = self.out_embeddings(context_idx)  # [batch_size, embed_dim]
        
        # Get embeddings for negative samples
        neg_emb = self.out_embeddings(neg_samples)  # [batch_size, neg_samples, embed_dim]
        
        # Compute positive score
        pos_score = torch.sum(target_emb * context_emb, dim=1)  # [batch_size]
        pos_score = F.logsigmoid(pos_score)  # [batch_size]
        
        # Compute negative score
        neg_score = torch.bmm(
            neg_emb, target_emb.unsqueeze(2)
        ).squeeze()  # [batch_size, neg_samples]
        neg_score = F.logsigmoid(-neg_score).sum(1)  # [batch_size]
        
        # Loss: maximize positive and minimize negative scores
        loss = -(pos_score + neg_score).mean()
        
        return loss
    
    def forward_cbow(self, context_indices, target_idx, neg_samples):
        """Forward pass for CBOW architecture"""
        # Get embeddings for context words and average them
        context_emb = self.in_embeddings(context_indices)  # [batch_size, context_size, embed_dim]
        context_emb = torch.mean(context_emb, dim=1)  # [batch_size, embed_dim]
        
        # Get embeddings for positive target word
        target_emb = self.out_embeddings(target_idx)  # [batch_size, embed_dim]
        
        # Get embeddings for negative samples
        neg_emb = self.out_embeddings(neg_samples)  # [batch_size, neg_samples, embed_dim]
        
        # Compute positive score
        pos_score = torch.sum(context_emb * target_emb, dim=1)  # [batch_size]
        pos_score = F.logsigmoid(pos_score)  # [batch_size]
        
        # Compute negative score
        neg_score = torch.bmm(
            neg_emb, context_emb.unsqueeze(2)
        ).squeeze()  # [batch_size, neg_samples]
        neg_score = F.logsigmoid(-neg_score).sum(1)  # [batch_size]
        
        # Loss: maximize positive and minimize negative scores
        loss = -(pos_score + neg_score).mean()
        
        return loss
    
    def forward(self, *args):
        """Forward pass selecting the appropriate architecture"""
        if self.mode == 'skipgram':
            return self.forward_skipgram(*args)
        else:  # CBOW
            return self.forward_cbow(*args)
    
    def get_word_vectors(self):
        """Return trained word vectors"""
        return self.in_embeddings.weight.data


class Word2VecTrainer:
    """Trainer for Word2Vec models"""
    
    def __init__(
        self,
        texts: List[List[str]],
        embedding_dim: int = 300,
        window_size: int = 5,
        mode: str = 'skipgram',
        min_count: int = 5,
        negative_samples: int = 5,
        ns_exponent: float = 0.75,
        subsampling_threshold: float = 1e-5,
        epochs: int = 5,
        batch_size: int = 32,
        learning_rate: float = 0.025,
        min_learning_rate: float = 0.0001,
        padding_idx: Optional[int] = None,
        sparse: bool = False
    ):
        """
        Initialize Word2Vec trainer
        
        Args:
            texts: List of tokenized texts
            embedding_dim: Dimension of word embeddings
            window_size: Context window size
            mode: Model architecture ('skipgram' or 'cbow')
            min_count: Minimum word frequency to include in vocabulary
            negative_samples: Number of negative samples per positive sample
            ns_exponent: Exponent for negative sampling distribution
            subsampling_threshold: Threshold for subsampling frequent words
            epochs: Number of training epochs
            batch_size: Training batch size
            learning_rate: Initial learning rate
            min_learning_rate: Minimum learning rate
            padding_idx: If specified, padded embeddings will remain zeroed
            sparse: If True, gradients for embeddings will be sparse
        """
        self.texts = texts
        self.embedding_dim = embedding_dim
        self.window_size = window_size
        self.mode = mode.lower()
        self.min_count = min_count
        self.negative_samples = negative_samples
        self.ns_exponent = ns_exponent
        self.subsampling_threshold = subsampling_threshold
        self.epochs = epochs
        self.batch_size = batch_size
        self.initial_lr = learning_rate
        self.min_lr = min_learning_rate
        self.padding_idx = padding_idx
        self.sparse = sparse
        
        # Build vocabulary
        self.word_to_idx = self._build_vocab()
        self.vocab_size = len(self.word_to_idx)
        
        # Create dataset
        self.dataset = Word2VecDataset(
            texts=texts,
            word_to_idx=self.word_to_idx,
            window_size=window_size,
            mode=mode,
            ns_exponent=ns_exponent,
            negative_samples=negative_samples,
            min_count=min_count,
            subsampling_threshold=subsampling_threshold
        )
        
        # Create model
        self.model = Word2Vec(
            vocab_size=self.vocab_size,
            embedding_dim=embedding_dim,
            mode=mode,
            padding_idx=padding_idx,
            sparse=sparse
        )
        
        # Create optimizer
        self.optimizer = optim.Adam(self.model.parameters(), lr=learning_rate)
        
        # Create dataloader
        self.dataloader = DataLoader(
            self.dataset,
            batch_size=batch_size,
            shuffle=True,
            num_workers=4,
            pin_memory=True
        )
    
    def _build_vocab(self) -> Dict[str, int]:
        """Build vocabulary from texts"""
        counter = Counter()
        for text in self.texts:
            counter.update(text)
        
        # Filter words by minimum count
        words = [word for word, count in counter.items() if count >= self.min_count]
        
        # Create word to index mapping
        word_to_idx = {word: i for i, word in enumerate(words)}
        
        logging.info(f"Vocabulary size: {len(word_to_idx)}")
        return word_to_idx
    
    def train(self):
        """Train the Word2Vec model"""
        logging.info(f"Training Word2Vec ({self.mode}) model")
        
        self.model.train()
        
        # Training loop
        total_steps = len(self.dataloader) * self.epochs
        progress_bar = tqdm(total=total_steps, desc="Training")
        
        for epoch in range(self.epochs):
            total_loss = 0
            
            # Adjust learning rate
            lr = max(self.min_lr, self.initial_lr * (1.0 - epoch / self.epochs))
            for param_group in self.optimizer.param_groups:
                param_group['lr'] = lr
            
            for batch in self.dataloader:
                self.optimizer.zero_grad()
                
                # Forward pass
                loss = self.model(*batch)
                
                # Backward pass
                loss.backward()
                self.optimizer.step()
                
                total_loss += loss.item()
                progress_bar.update(1)
            
            avg_loss = total_loss / len(self.dataloader)
            logging.info(f"Epoch {epoch+1}/{self.epochs}, Loss: {avg_loss:.4f}, LR: {lr:.6f}")
        
        progress_bar.close()
        logging.info("Training complete")
    
    def get_word_vectors(self) -> Dict[str, torch.Tensor]:
        """Get word vectors as a dictionary mapping words to embeddings"""
        self.model.eval()
        vectors = self.model.get_word_vectors()
        
        word_vectors = {}
        for word, idx in self.word_to_idx.items():
            word_vectors[word] = vectors[idx].cpu().numpy()
            
        return word_vectors
    
    def save_vectors(self, file_path: str):
        """Save word vectors in word2vec format"""
        vectors = self.get_word_vectors()
        
        with open(file_path, 'w', encoding='utf-8') as f:
            f.write(f"{len(vectors)} {self.embedding_dim}\n")
            for word, vec in vectors.items():
                vec_str = ' '.join(map(str, vec))
                f.write(f"{word} {vec_str}\n")
        
        logging.info(f"Word vectors saved to {file_path}")
    
    def save_model(self, file_path: str):
        """Save model parameters"""
        torch.save({
            'model_state_dict': self.model.state_dict(),
            'word_to_idx': self.word_to_idx,
            'embedding_dim': self.embedding_dim,
            'mode': self.mode
        }, file_path)
        
        logging.info(f"Model saved to {file_path}")
    
    @classmethod
    def load_model(cls, file_path: str):
        """Load a pre-trained model"""
        checkpoint = torch.load(file_path)
        
        # Initialize an empty trainer
        trainer = cls.__new__(cls)
        
        # Set attributes from checkpoint
        trainer.word_to_idx = checkpoint['word_to_idx']
        trainer.vocab_size = len(trainer.word_to_idx)
        trainer.embedding_dim = checkpoint['embedding_dim']
        trainer.mode = checkpoint['mode']
        
        # Create model and load weights
        trainer.model = Word2Vec(
            vocab_size=trainer.vocab_size,
            embedding_dim=trainer.embedding_dim,
            mode=trainer.mode
        )
        trainer.model.load_state_dict(checkpoint['model_state_dict'])
        
        return trainer


def train_word2vec(
    texts: List[List[str]],
    embedding_dim: int = 300,
    window_size: int = 5,
    mode: str = 'skipgram',
    min_count: int = 5,
    negative_samples: int = 5,
    epochs: int = 5,
    batch_size: int = 32,
    learning_rate: float = 0.025,
    save_path: Optional[str] = None
) -> Word2VecTrainer:
    """
    Train a Word2Vec model and return the trainer
    
    Args:
        texts: List of tokenized texts
        embedding_dim: Dimension of word embeddings
        window_size: Context window size
        mode: Model architecture ('skipgram' or 'cbow')
        min_count: Minimum word frequency to include in vocabulary
        negative_samples: Number of negative samples per positive sample
        epochs: Number of training epochs
        batch_size: Training batch size
        learning_rate: Initial learning rate
        save_path: If provided, save the trained model to this path
        
    Returns:
        Trained Word2VecTrainer instance
    """
    trainer = Word2VecTrainer(
        texts=texts,
        embedding_dim=embedding_dim,
        window_size=window_size,
        mode=mode,
        min_count=min_count,
        negative_samples=negative_samples,
        epochs=epochs,
        batch_size=batch_size,
        learning_rate=learning_rate
    )
    
    trainer.train()
    
    if save_path:
        trainer.save_model(save_path)
    
    return trainer



"""
Useful pandas operations for data manipulation and analysis.
This module provides examples of common pandas operations including:
- Data creation and loading
- Data exploration
- Filtering and selection
- Grouping and aggregation
- Handling missing values
- Merging and reshaping
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from io import StringIO

def create_sample_dataframe():
    """Create a sample DataFrame for demonstration."""
    df = pd.DataFrame({
        'A': np.random.rand(10),
        'B': np.random.randint(0, 10, 10),
        'C': pd.date_range('2023-01-01', periods=10),
        'D': pd.Categorical(['X', 'Y', 'Z', 'X', 'Y', 'Z', 'X', 'Y', 'Z', 'X']),
        'E': ['alpha', 'beta', 'gamma', 'delta', 'epsilon', 
              'zeta', 'eta', 'theta', 'iota', 'kappa']
    })
    # Add some missing values
    df.loc[2:4, 'A'] = np.nan
    df.loc[7, 'E'] = None
    return df

def basic_exploration(df):
    """Demonstrate basic dataframe exploration methods."""
    print("=== Basic DataFrame Exploration ===")
    print("\nFirst 5 rows:")
    print(df.head())
    
    print("\nDataFrame information:")
    print(df.info())
    
    print("\nSummary statistics:")
    print(df.describe(include='all'))
    
    print("\nColumn data types:")
    print(df.dtypes)
    
    print("\nCheck for missing values:")
    print(df.isna().sum())
    
    return {
        'shape': df.shape,
        'columns': df.columns.tolist(),
        'missing_values': df.isna().sum().to_dict()
    }

def filtering_and_selection(df):
    """Demonstrate filtering and selection operations."""
    print("=== Filtering and Selection ===")
    
    print("\nSelect column 'A':")
    print(df['A'])
    
    print("\nSelect multiple columns:")
    print(df[['A', 'B', 'E']])
    
    print("\nFilter rows where B > 5:")
    print(df[df['B'] > 5])
    
    print("\nFilter rows with complex condition (B > 5 and D == 'X'):")
    print(df[(df['B'] > 5) & (df['D'] == 'X')])
    
    print("\nSelect using .loc (label-based):")
    print(df.loc[2:4, ['A', 'B', 'E']])
    
    print("\nSelect using .iloc (position-based):")
    print(df.iloc[2:4, [0, 1, 4]])
    
    print("\nBoolean indexing with .loc:")
    print(df.loc[df['D'] == 'X', 'B'])

def handle_missing_data(df):
    """Demonstrate methods for handling missing data."""
    print("=== Handling Missing Data ===")
    
    print("\nDrop rows with any NaN values:")
    print(df.dropna())
    
    print("\nDrop rows with all NaN values:")
    print(df.dropna(how='all'))
    
    print("\nFill NaN values with 0:")
    print(df.fillna(0))
    
    print("\nFill NaN values with column mean:")
    print(df.fillna(df.mean(numeric_only=True)))
    
    print("\nFill NaN with different values per column:")
    print(df.fillna({'A': 0, 'E': 'Unknown'}))
    
    print("\nInterpolate missing values:")
    print(df.interpolate())

def grouping_and_aggregation(df):
    """Demonstrate grouping and aggregation operations."""
    print("=== Grouping and Aggregation ===")
    
    print("\nGroup by column D and get mean of numeric columns:")
    print(df.groupby('D').mean(numeric_only=True))
    
    print("\nGroup by column D and get multiple aggregations:")
    print(df.groupby('D').agg({'A': ['mean', 'std'], 'B': ['min', 'max']}))
    
    print("\nGroup by multiple columns:")
    print(df.groupby(['D', df['B'] > 5]).mean(numeric_only=True))
    
    print("\nGroup by with custom aggregation function:")
    print(df.groupby('D')['B'].agg(lambda x: x.max() - x.min()))
    
    print("\nPivot table:")
    pivot_df = pd.DataFrame({
        'date': pd.date_range('2023-01-01', periods=20),
        'category': np.random.choice(['X', 'Y', 'Z'], 20),
        'product': np.random.choice(['P1', 'P2', 'P3', 'P4'], 20),
        'sales': np.random.randint(10, 100, 20)
    })
    print(pd.pivot_table(pivot_df, values='sales', index='category', 
                        columns='product', aggfunc='sum', fill_value=0))

def merging_and_joining(df):
    """Demonstrate merging and joining dataframes."""
    print("=== Merging and Joining DataFrames ===")
    
    # Create two sample dataframes
    df1 = pd.DataFrame({
        'key': ['A', 'B', 'C', 'D'],
        'value1': [1, 2, 3, 4]
    })
    
    df2 = pd.DataFrame({
        'key': ['B', 'D', 'E', 'F'],
        'value2': [5, 6, 7, 8]
    })
    
    print("\nInner join:")
    print(pd.merge(df1, df2, on='key'))
    
    print("\nLeft join:")
    print(pd.merge(df1, df2, on='key', how='left'))
    
    print("\nRight join:")
    print(pd.merge(df1, df2, on='key', how='right'))
    
    print("\nOuter join:")
    print(pd.merge(df1, df2, on='key', how='outer'))
    
    print("\nConcatenate vertically:")
    print(pd.concat([df1, df1]))
    
    print("\nConcatenate horizontally:")
    print(pd.concat([df1, df2.drop('key', axis=1)], axis=1))

def data_transformation(df):
    """Demonstrate data transformation operations."""
    print("=== Data Transformation ===")
    
    print("\nApply function to each element:")
    print(df['B'].apply(lambda x: x ** 2))
    
    print("\nMap values to new values:")
    print(df['D'].map({'X': 1, 'Y': 2, 'Z': 3}))
    
    print("\nOne-hot encoding categorical column:")
    print(pd.get_dummies(df['D']))
    
    print("\nBin continuous data into categories:")
    print(pd.cut(df['B'], bins=3, labels=['Small', 'Medium', 'Large']))
    
    print("\nReshape with melt (wide to long):")
    melted = pd.melt(df[['D', 'A', 'B']], id_vars=['D'])
    print(melted)
    
    print("\nPivot (long to wide):")
    print(melted.pivot(index='D', columns='variable', values='value'))

def time_series_operations():
    """Demonstrate time series operations."""
    print("=== Time Series Operations ===")
    
    # Create a time series
    dates = pd.date_range('2023-01-01', periods=100)
    ts = pd.Series(np.random.randn(len(dates)).cumsum(), index=dates)
    
    print("\nResample to monthly frequency (downsample):")
    print(ts.resample('M').mean())
    
    print("\nResample to 2-day frequency (downsample):")
    print(ts.resample('2D').mean())
    
    print("\nShift time series forward by 2 periods:")
    print(ts.shift(2).head())
    
    print("\nPercent change between periods:")
    print(ts.pct_change().head())
    
    print("\nRolling window calculations (7-day moving average):")
    print(ts.rolling(7).mean().head(10))
    
    print("\nExpanding window calculations (cumulative mean):")
    print(ts.expanding().mean().head(10))

def io_operations():
    """Demonstrate I/O operations with pandas."""
    print("=== I/O Operations ===")
    
    df = create_sample_dataframe()
    
    print("\nWrite to CSV:")
    df.to_csv("sample.csv")
    print("DataFrame written to 'sample.csv'")
    
    print("\nRead from CSV:")
    df_read = pd.read_csv("sample.csv", index_col=0)
    print(df_read.head())
    
    print("\nWrite to Excel:")
    df.to_excel("sample.xlsx", sheet_name='Sheet1')
    print("DataFrame written to 'sample.xlsx'")
    
    print("\nCSV from string:")
    csv_data = """
    A,B,C
    1,2,3
    4,5,6
    7,8,9
    """
    df_csv = pd.read_csv(StringIO(csv_data.strip()))
    print(df_csv)
    
    print("\nRead from JSON:")
    df_json = pd.read_json(df.to_json(orient='records'))
    print(df_json.head())

def main():
    """Run all pandas examples."""
    df = create_sample_dataframe()
    
    # Basic operations
    basic_exploration(df)
    filtering_and_selection(df)
    handle_missing_data(df)
    grouping_and_aggregation(df)
    merging_and_joining(df)
    data_transformation(df)
    time_series_operations()
    io_operations()
    
    print("\nPlotting example:")
    df.plot(x='C', y='A', kind='line', figsize=(10, 6))
    plt.title('Sample Time Series Plot')
    plt.tight_layout()
    plt.show()

if __name__ == "__main__":
    main()



import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import Dataset, DataLoader
import numpy as np
from collections import Counter
import re

class TextDataset(Dataset):
    def __init__(self, texts, labels, vocab=None, max_length=100):
        self.texts = texts
        self.labels = labels
        self.max_length = max_length
        
        # Build vocabulary if not provided
        if vocab is None:
            self.build_vocab()
        else:
            self.vocab = vocab
            
    def build_vocab(self):
        # Tokenize all texts
        all_tokens = []
        for text in self.texts:
            all_tokens.extend(self.tokenize(text))
            
        # Count token frequencies
        counter = Counter(all_tokens)
        vocab_tokens = ["<PAD>", "<UNK>"] + [token for token, _ in counter.most_common(5000)]
        
        # Create vocabulary mapping
        self.vocab = {token: idx for idx, token in enumerate(vocab_tokens)}
        
    def tokenize(self, text):
        # Simple tokenization
        text = text.lower()
        text = re.sub(r'[^\w\s]', '', text)
        return text.split()
    
    def __len__(self):
        return len(self.texts)
    
    def __getitem__(self, idx):
        text = self.texts[idx]
        label = self.labels[idx]
        
        # Tokenize and convert to indices
        tokens = self.tokenize(text)
        token_ids = [self.vocab.get(token, 1) for token in tokens]  # 1 is <UNK>
        
        # Truncate or pad to max_length
        if len(token_ids) > self.max_length:
            token_ids = token_ids[:self.max_length]
        else:
            token_ids = token_ids + [0] * (self.max_length - len(token_ids))  # 0 is <PAD>
            
        return torch.tensor(token_ids), torch.tensor(label)

class BiLSTMClassifier(nn.Module):
    def __init__(self, vocab_size, embedding_dim, hidden_dim, output_dim):
        super(BiLSTMClassifier, self).__init__()
        self.embedding = nn.Embedding(vocab_size, embedding_dim, padding_idx=0)
        self.lstm = nn.LSTM(embedding_dim, hidden_dim, batch_first=True, bidirectional=True)
        self.fc = nn.Linear(hidden_dim * 2, output_dim)
        self.dropout = nn.Dropout(0.5)
        
    def forward(self, text):
        # text: [batch_size, seq_len]
        embedded = self.dropout(self.embedding(text))  # [batch_size, seq_len, emb_dim]
        
        # Pass through LSTM
        output, (hidden, _) = self.lstm(embedded)
        # hidden: [2, batch_size, hidden_dim]
        
        # Concat the final forward and backward hidden states
        hidden = torch.cat((hidden[-2,:,:], hidden[-1,:,:]), dim=1)
        # hidden: [batch_size, hidden_dim*2]
        
        return self.fc(self.dropout(hidden))

def train(model, train_loader, epochs=5):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    model = model.to(device)
    optimizer = optim.Adam(model.parameters())
    criterion = nn.CrossEntropyLoss()
    
    for epoch in range(epochs):
        model.train()
        total_loss = 0
        
        for texts, labels in train_loader:
            texts, labels = texts.to(device), labels.to(device)
            
            optimizer.zero_grad()
            predictions = model(texts)
            loss = criterion(predictions, labels)
            loss.backward()
            optimizer.step()
            
            total_loss += loss.item()
            
        print(f'Epoch {epoch+1}, Loss: {total_loss/len(train_loader):.4f}')

def evaluate(model, test_loader):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    model = model.to(device)
    model.eval()
    
    correct = 0
    total = 0
    
    with torch.no_grad():
        for texts, labels in test_loader:
            texts, labels = texts.to(device), labels.to(device)
            predictions = model(texts)
            _, predicted = torch.max(predictions, 1)
            total += labels.size(0)
            correct += (predicted == labels).sum().item()
    
    accuracy = correct / total
    print(f'Test Accuracy: {accuracy * 100:.2f}%')
    return accuracy

def generate_data():
    # Generate sample text data for demonstration
    texts = [
        "this movie was great and entertaining",
        "really enjoyed the storyline",
        "terrible acting and boring plot",
        "waste of time and money",
        "best film I have seen this year",
        # Add more examples...
    ]
    labels = [1, 1, 0, 0, 1]  # 1: positive, 0: negative
    
    # Split into train/test
    train_texts = texts[:4]
    train_labels = labels[:4]
    test_texts = texts[4:]
    test_labels = labels[4:]
    
    return train_texts, train_labels, test_texts, test_labels

def main():
    # In a real scenario, you would load your dataset
    train_texts, train_labels, test_texts, test_labels = generate_data()
    
    # Create datasets
    train_dataset = TextDataset(train_texts, train_labels)
    test_dataset = TextDataset(test_texts, test_labels, vocab=train_dataset.vocab)
    
    # Create data loaders
    train_loader = DataLoader(train_dataset, batch_size=2, shuffle=True)
    test_loader = DataLoader(test_dataset, batch_size=2)
    
    # Initialize model
    vocab_size = len(train_dataset.vocab)
    embedding_dim = 100
    hidden_dim = 128
    output_dim = 2  # Binary classification
    
    model = BiLSTMClassifier(vocab_size, embedding_dim, hidden_dim, output_dim)
    
    # Train and evaluate
    train(model, train_loader, epochs=5)
    evaluate(model, test_loader)

if __name__ == '__main__':
    main()



"""
Tokenizer for text processing optimized for word embedding models like Word2Vec and GloVe.
"""

import re
import string
import unicodedata
from typing import List, Dict, Set, Tuple, Optional, Callable, Union, Iterator, Any
import logging
from collections import Counter
import nltk
from nltk.tokenize import word_tokenize, sent_tokenize
from nltk.stem import PorterStemmer, WordNetLemmatizer
from nltk.corpus import stopwords

# Add PyTorch import
try:
    import torch
    TORCH_AVAILABLE = True
except ImportError:
    TORCH_AVAILABLE = False

# Ensure NLTK resources are available
try:
    nltk.data.find('tokenizers/punkt')
except LookupError:
    nltk.download('punkt')
try:
    nltk.data.find('corpora/stopwords')
except LookupError:
    nltk.download('stopwords')
try:
    nltk.data.find('corpora/wordnet')
except LookupError:
    nltk.download('wordnet')


class Tokenizer:
    """Flexible tokenizer for text preprocessing and tokenization."""
    
    def __init__(
        self,
        lowercase: bool = True,
        remove_punctuation: bool = True,
        remove_digits: bool = False,
        remove_stopwords: bool = False,
        stem: bool = False,
        lemmatize: bool = False,
        min_token_length: int = 1,
        max_token_length: Optional[int] = None,
        language: str = 'english',
        custom_stopwords: Optional[Set[str]] = None,
        custom_filters: Optional[List[Callable[[str], str]]] = None,
        keep_n_grams: Optional[List[int]] = None,
        n_gram_delimiter: str = '_',
        # New parameters for special tokens
        pad_token: str = "<pad>",
        unk_token: str = "<unk>",
        bos_token: str = "<bos>",
        eos_token: str = "<eos>",
        mask_token: str = "<mask>",
        special_tokens: Optional[List[str]] = None
    ):
        """
        Initialize the tokenizer with specified preprocessing options.
        
        Args:
            lowercase: Convert text to lowercase
            remove_punctuation: Remove punctuation
            remove_digits: Remove numeric digits
            remove_stopwords: Remove common stopwords
            stem: Apply stemming (Porter stemmer)
            lemmatize: Apply lemmatization (WordNet lemmatizer)
            min_token_length: Minimum length for a token to be kept
            max_token_length: Maximum length for a token to be kept (None = no limit)
            language: Language for stopwords and other language-specific processing
            custom_stopwords: Additional stopwords to remove
            custom_filters: List of custom filter functions for additional preprocessing
            keep_n_grams: List of n-gram sizes to generate (e.g., [2, 3] for bigrams and trigrams)
            n_gram_delimiter: Character to join n-gram components
            pad_token: Token used for padding sequences to equal length
            unk_token: Token used for unknown words
            bos_token: Beginning of sequence token
            eos_token: End of sequence token
            mask_token: Token used for masked language modeling
            special_tokens: Additional special tokens to include
        """
        self.lowercase = lowercase
        self.remove_punctuation = remove_punctuation
        self.remove_digits = remove_digits
        self.remove_stopwords = remove_stopwords
        self.stem = stem
        self.lemmatize = lemmatize
        self.min_token_length = min_token_length
        self.max_token_length = max_token_length
        self.language = language
        self.custom_filters = custom_filters or []
        self.keep_n_grams = keep_n_grams or []
        self.n_gram_delimiter = n_gram_delimiter
        
        # Set up stopwords
        if remove_stopwords:
            self.stopwords = set(stopwords.words(language))
            if custom_stopwords:
                self.stopwords.update(custom_stopwords)
        else:
            self.stopwords = set()
            if custom_stopwords:
                self.stopwords = custom_stopwords
        
        # Set up stemmers and lemmatizers if needed
        if stem:
            self.stemmer = PorterStemmer()
        
        if lemmatize:
            self.lemmatizer = WordNetLemmatizer()
        
        # Set up punctuation translation table
        if remove_punctuation:
            self.punctuation_table = str.maketrans('', '', string.punctuation)
        
        # Set up digit translation table
        if remove_digits:
            self.digit_table = str.maketrans('', '', string.digits)
        
        # Initialize special tokens
        self.pad_token = pad_token
        self.unk_token = unk_token
        self.bos_token = bos_token
        self.eos_token = eos_token
        self.mask_token = mask_token
        self.default_special_tokens = [pad_token, unk_token, bos_token, eos_token, mask_token]
        self.special_tokens = self.default_special_tokens + (special_tokens or [])
        
        # Dictionary to store token-to-id and id-to-token mappings
        self.token_to_id = {}
        self.id_to_token = {}
        
        # For handling unknown tokens
        self.unk_token_id = None
        self.pad_token_id = None
    
    def _preprocess_text(self, text: str) -> str:
        """Apply preprocessing steps to text."""
        # Apply lowercasing
        if self.lowercase:
            text = text.lower()
        
        # Apply custom filters
        for filter_fn in self.custom_filters:
            text = filter_fn(text)
        
        # Remove punctuation
        if self.remove_punctuation:
            text = text.translate(self.punctuation_table)
        
        # Remove digits
        if self.remove_digits:
            text = text.translate(self.digit_table)
        
        return text
    
    def _process_token(self, token: str) -> str:
        """Apply token-level processing (stemming, lemmatization)."""
        # Apply stemming
        if self.stem:
            token = self.stemmer.stem(token)
        
        # Apply lemmatization
        if self.lemmatize:
            token = self.lemmatizer.lemmatize(token)
        
        return token
    
    def _is_valid_token(self, token: str) -> bool:
        """Check if token meets criteria for inclusion."""
        # Check minimum length
        if len(token) < self.min_token_length:
            return False
        
        # Check maximum length
        if self.max_token_length and len(token) > self.max_token_length:
            return False
        
        # Check if token is a stopword
        if token in self.stopwords:
            return False
        
        # Remove tokens that are just whitespace
        if token.strip() == '':
            return False
        
        return True
    
    def _generate_n_grams(self, tokens: List[str]) -> List[str]:
        """Generate n-grams from tokens."""
        result = tokens.copy()
        
        for n in self.keep_n_grams:
            if n < 2 or n > len(tokens):
                continue
                
            for i in range(len(tokens) - n + 1):
                n_gram = self.n_gram_delimiter.join(tokens[i:i+n])
                result.append(n_gram)
        
        return result
    
    def tokenize_text(self, text: str) -> List[str]:
        """
        Tokenize a single text string according to configured options.
        
        Args:
            text: Input text to tokenize
            
        Returns:
            List of processed tokens
        """
        # Normalize unicode
        text = unicodedata.normalize('NFKD', text)
        
        # Apply preprocessing
        text = self._preprocess_text(text)
        
        # Tokenize into words
        tokens = word_tokenize(text, language=self.language)
        
        # Process tokens
        processed_tokens = []
        for token in tokens:
            # Apply token-level processing
            processed = self._process_token(token)
            
            # Check if token is valid
            if self._is_valid_token(processed):
                processed_tokens.append(processed)
        
        # Generate n-grams if specified
        if self.keep_n_grams:
            processed_tokens = self._generate_n_grams(processed_tokens)
        
        return processed_tokens
    
    def tokenize_texts(self, texts: List[str]) -> List[List[str]]:
        """
        Tokenize a list of texts.
        
        Args:
            texts: List of texts to tokenize
            
        Returns:
            List of lists of tokens
        """
        return [self.tokenize_text(text) for text in texts]
    
    def tokenize_into_sentences(self, text: str) -> List[List[str]]:
        """
        Tokenize text into sentences, then tokenize each sentence into words.
        
        Args:
            text: Input text
            
        Returns:
            List of lists of tokens, where each inner list represents a sentence
        """
        # Split into sentences
        sentences = sent_tokenize(text, language=self.language)
        
        # Tokenize each sentence
        return self.tokenize_texts(sentences)
    
    def build_vocab(
        self, 
        texts: List[str], 
        min_count: int = 5, 
        max_vocab_size: Optional[int] = None
    ) -> Dict[str, int]:
        """
        Build vocabulary from texts.
        
        Args:
            texts: List of texts to process
            min_count: Minimum occurrence count to include a token
            max_vocab_size: Maximum vocabulary size (None = no limit)
            
        Returns:
            Dictionary mapping tokens to their indices
        """
        # Tokenize all texts
        all_tokens = []
        for text in texts:
            all_tokens.extend(self.tokenize_text(text))
        
        # Count tokens
        counter = Counter(all_tokens)
        
        # Filter by min_count
        filtered_tokens = [(token, count) for token, count in counter.items() 
                           if count >= min_count]
        
        # Sort by frequency (most frequent first)
        sorted_tokens = sorted(filtered_tokens, key=lambda x: x[1], reverse=True)
        
        # Limit vocab size if specified, reserving space for special tokens
        if max_vocab_size:
            available_space = max_vocab_size - len(self.special_tokens)
            sorted_tokens = sorted_tokens[:available_space]
        
        # Create word to index mapping, giving priority to special tokens
        vocab = {}
        idx = 0
        
        # First add special tokens
        for token in self.special_tokens:
            vocab[token] = idx
            idx += 1
            
        # Then add the rest of the tokens
        for token, _ in sorted_tokens:
            if token not in vocab:  # Skip if token is already in vocab (e.g., as a special token)
                vocab[token] = idx
                idx += 1
        
        # Store the mappings
        self.token_to_id = vocab
        self.id_to_token = {idx: token for token, idx in vocab.items()}
        
        # Store special token IDs for convenience
        self.pad_token_id = vocab.get(self.pad_token)
        self.unk_token_id = vocab.get(self.unk_token)
        
        return vocab
    
    def prepare_for_word_embeddings(
        self, 
        texts: List[str],
        min_count: int = 5,
        max_vocab_size: Optional[int] = None
    ) -> Tuple[List[List[str]], Dict[str, int]]:
        """
        Prepare texts for word embedding models like Word2Vec or GloVe.
        
        Args:
            texts: List of raw text strings
            min_count: Minimum occurrence count to include a token
            max_vocab_size: Maximum vocabulary size (None = no limit)
            
        Returns:
            Tuple of (tokenized_texts, vocabulary)
        """
        # Tokenize texts
        tokenized_texts = self.tokenize_texts(texts)
        
        # Count all tokens
        counter = Counter()
        for tokens in tokenized_texts:
            counter.update(tokens)
        
        # Filter by min_count
        filtered_tokens = [(token, count) for token, count in counter.items() 
                          if count >= min_count]
        
        # Sort by frequency
        sorted_tokens = sorted(filtered_tokens, key=lambda x: x[1], reverse=True)
        
        # Limit vocab size if specified, reserving space for special tokens
        if max_vocab_size:
            available_space = max_vocab_size - len(self.special_tokens)
            sorted_tokens = sorted_tokens[:available_space]
        
        # Create vocab, giving priority to special tokens
        vocab = {}
        idx = 0
        
        # First add special tokens
        for token in self.special_tokens:
            vocab[token] = idx
            idx += 1
            
        # Then add the rest of the tokens
        for token, _ in sorted_tokens:
            if token not in vocab:  # Skip if token is already in vocab
                vocab[token] = idx
                idx += 1
        
        # Store the mappings
        self.token_to_id = vocab
        self.id_to_token = {idx: token for token, idx in vocab.items()}
        
        # Store special token IDs for convenience
        self.pad_token_id = vocab.get(self.pad_token)
        self.unk_token_id = vocab.get(self.unk_token)
        
        return tokenized_texts, vocab
    
    # New token conversion methods
    def convert_tokens_to_ids(self, tokens: List[str]) -> List[int]:
        """
        Convert a list of tokens to their corresponding IDs.
        
        Args:
            tokens: List of tokens to convert
            
        Returns:
            List of token IDs
        """
        if not self.token_to_id:
            raise ValueError("Vocabulary not built yet. Call build_vocab() or prepare_for_word_embeddings() first.")
            
        return [self.token_to_id.get(token, self.unk_token_id) for token in tokens]
    
    def convert_ids_to_tokens(self, ids: List[int]) -> List[str]:
        """
        Convert a list of token IDs back to tokens.
        
        Args:
            ids: List of token IDs to convert
            
        Returns:
            List of tokens
        """
        if not self.id_to_token:
            raise ValueError("Vocabulary not built yet. Call build_vocab() or prepare_for_word_embeddings() first.")
            
        return [self.id_to_token.get(idx, self.unk_token) for idx in ids]
    
    def encode(self, text: str, add_special_tokens: bool = False) -> List[int]:
        """
        Tokenize and convert text to token IDs.
        
        Args:
            text: Text to encode
            add_special_tokens: Whether to add BOS/EOS tokens
            
        Returns:
            List of token IDs
        """
        tokens = self.tokenize_text(text)
        
        if add_special_tokens:
            tokens = [self.bos_token] + tokens + [self.eos_token]
            
        return self.convert_tokens_to_ids(tokens)
    
    def decode(self, ids: List[int], skip_special_tokens: bool = False) -> str:
        """
        Convert token IDs back to text.
        
        Args:
            ids: List of token IDs
            skip_special_tokens: Whether to skip special tokens in output
            
        Returns:
            Decoded text
        """
        tokens = self.convert_ids_to_tokens(ids)
        
        if skip_special_tokens:
            tokens = [token for token in tokens if token not in self.special_tokens]
            
        return " ".join(tokens)
    
    def batch_encode(self, texts: List[str], add_special_tokens: bool = False) -> List[List[int]]:
        """
        Encode a batch of texts to token IDs.
        
        Args:
            texts: List of texts to encode
            add_special_tokens: Whether to add BOS/EOS tokens
            
        Returns:
            List of lists of token IDs
        """
        return [self.encode(text, add_special_tokens) for text in texts]
    
    # PyTorch integration methods
    def encode_as_tensor(
        self, 
        text: str, 
        add_special_tokens: bool = False, 
        return_tensors: str = 'pt'
    ) -> Union[List[int], 'torch.Tensor']:
        """
        Encode text and return as PyTorch tensor if requested.
        
        Args:
            text: Text to encode
            add_special_tokens: Whether to add BOS/EOS tokens
            return_tensors: Output format ('pt' for PyTorch tensor, None for list)
            
        Returns:
            Token IDs as list or PyTorch tensor
        """
        ids = self.encode(text, add_special_tokens)
        
        if return_tensors == 'pt':
            if not TORCH_AVAILABLE:
                raise ImportError("PyTorch is not installed. Install it with 'pip install torch'.")
            return torch.tensor(ids)
        
        return ids
    
    def batch_encode_as_tensors(
        self, 
        texts: List[str], 
        add_special_tokens: bool = False,
        padding: bool = False,
        truncation: bool = False,
        max_length: Optional[int] = None,
        return_attention_mask: bool = False,
        return_tensors: str = 'pt'
    ) -> Dict[str, Union[List, 'torch.Tensor']]:
        """
        Encode a batch of texts with padding and attention masks.
        
        Args:
            texts: List of texts to encode
            add_special_tokens: Whether to add BOS/EOS tokens
            padding: Whether to pad sequences to equal length
            truncation: Whether to truncate sequences longer than max_length
            max_length: Maximum sequence length (if truncation is True)
            return_attention_mask: Whether to return attention mask
            return_tensors: Output format ('pt' for PyTorch tensor, None for list)
            
        Returns:
            Dictionary with 'input_ids' and optionally 'attention_mask'
        """
        # Encode all texts
        encoded_texts = [self.encode(text, add_special_tokens) for text in texts]
        
        # Handle truncation if requested
        if truncation and max_length:
            encoded_texts = [ids[:max_length] for ids in encoded_texts]
        
        # Get the length of each sequence for padding and attention mask
        seq_lengths = [len(ids) for ids in encoded_texts]
        
        # Handle padding if requested
        if padding:
            # Determine padding length
            if max_length:
                padding_length = max_length
            else:
                padding_length = max(seq_lengths)
            
            # Pad sequences
            padded_texts = []
            attention_masks = []
            
            for ids in encoded_texts:
                # Calculate padding needed
                padding_needed = padding_length - len(ids)
                
                # Pad with pad token ID
                padded_ids = ids + [self.pad_token_id] * padding_needed
                padded_texts.append(padded_ids)
                
                # Create attention mask (1 for actual tokens, 0 for padding)
                if return_attention_mask:
                    mask = [1] * len(ids) + [0] * padding_needed
                    attention_masks.append(mask)
        else:
            padded_texts = encoded_texts
            attention_masks = [[1] * len(ids) for ids in encoded_texts] if return_attention_mask else None
        
        # Prepare output dictionary
        output = {'input_ids': padded_texts}
        if return_attention_mask:
            output['attention_mask'] = attention_masks
            
        # Convert to tensors if requested
        if return_tensors == 'pt':
            if not TORCH_AVAILABLE:
                raise ImportError("PyTorch is not installed. Install it with 'pip install torch'.")
            
            for key in output:
                output[key] = torch.tensor(output[key])
                
        return output
    
    @staticmethod
    def collate_fn(batch: List[Dict[str, Any]], 
                  pad_token_id: int) -> Dict[str, 'torch.Tensor']:
        """
        Collate function for PyTorch DataLoader.
        
        Args:
            batch: List of dictionaries with 'input_ids' and other fields
            pad_token_id: ID of the padding token
            
        Returns:
            Dictionary with batched tensors
        """
        if not TORCH_AVAILABLE:
            raise ImportError("PyTorch is not installed. Install it with 'pip install torch'.")
            
        # Get all keys from the first batch element
        keys = batch[0].keys()
        
        result = {}
        for key in keys:
            if key == 'input_ids':
                # Get sequence lengths
                lengths = [len(item[key]) for item in batch]
                max_length = max(lengths)
                
                # Pad sequences and create tensor
                padded = [
                    item[key] + [pad_token_id] * (max_length - len(item[key]))
                    for item in batch
                ]
                result[key] = torch.tensor(padded)
                
                # Add sequence lengths if not present
                if 'lengths' not in keys:
                    result['lengths'] = torch.tensor(lengths)
                    
            elif key == 'attention_mask':
                # Get sequence lengths
                lengths = [len(item['input_ids']) for item in batch]
                max_length = max(lengths)
                
                # Create attention masks
                attention_masks = [
                    [1] * len(item['input_ids']) + [0] * (max_length - len(item['input_ids']))
                    for item in batch
                ]
                result[key] = torch.tensor(attention_masks)
                
            elif isinstance(batch[0][key], torch.Tensor):
                # For other tensor fields, stack them
                result[key] = torch.stack([item[key] for item in batch])
                
            elif isinstance(batch[0][key], (int, float, bool)):
                # For scalar fields, create a tensor
                result[key] = torch.tensor([item[key] for item in batch])
                
            else:
                # For other fields, just collect them as a list
                result[key] = [item[key] for item in batch]
                
        return result


# Helper functions for common text cleaning tasks
def clean_html(text: str) -> str:
    """Remove HTML tags from text."""
    html_pattern = re.compile('<.*?>')
    return html_pattern.sub('', text)

def clean_urls(text: str) -> str:
    """Remove URLs from text."""
    url_pattern = re.compile(r'https?://\S+|www\.\S+')
    return url_pattern.sub('', text)

def clean_emojis(text: str) -> str:
    """Remove emojis from text."""
    emoji_pattern = re.compile(
        "["
        "\U0001F600-\U0001F64F"  # emoticons
        "\U0001F300-\U0001F5FF"  # symbols & pictographs
        "\U0001F680-\U0001F6FF"  # transport & map symbols
        "\U0001F700-\U0001F77F"  # alchemical symbols
        "\U0001F780-\U0001F7FF"  # Geometric Shapes
        "\U0001F800-\U0001F8FF"  # Supplemental Arrows-C
        "\U0001F900-\U0001F9FF"  # Supplemental Symbols and Pictographs
        "\U0001FA00-\U0001FA6F"  # Chess Symbols
        "\U0001FA70-\U0001FAFF"  # Symbols and Pictographs Extended-A
        "\U00002702-\U000027B0"  # Dingbats
        "\U000024C2-\U0001F251" 
        "]+"
    )
    return emoji_pattern.sub('', text)

def clean_multiple_spaces(text: str) -> str:
    """Replace multiple spaces with single space."""
    return re.sub(r'\s+', ' ', text).strip()


# Example usage
def example_usage():
    texts = [
        "Hello world! This is an example sentence for tokenization.",
        "Word2Vec and GloVe are popular word embedding models.",
        "NLP tasks often require preprocessing steps like stemming and lemmatization.",
        "The quick brown fox jumps over the lazy dog.",
    ]
    
    # Basic tokenizer
    tokenizer = Tokenizer(
        lowercase=True,
        remove_punctuation=True,
        remove_stopwords=True,
        language='english'
    )
    
    # Tokenize texts
    tokenized_texts = tokenizer.tokenize_texts(texts)
    print(f"Tokenized texts: {tokenized_texts}")
    
    # Prepare for word embeddings
    processed_texts, vocab = tokenizer.prepare_for_word_embeddings(texts, min_count=1)
    print(f"Vocabulary size: {len(vocab)}")
    print(f"Vocabulary: {vocab}")
    
    # Demonstrate tokenizing with n-grams
    n_gram_tokenizer = Tokenizer(
        lowercase=True,
        remove_punctuation=True,
        keep_n_grams=[2, 3]
    )
    
    tokenized_with_ngrams = n_gram_tokenizer.tokenize_text(texts[0])
    print(f"With n-grams: {tokenized_with_ngrams}")
    
    # Example with special tokens and PyTorch integration
    if TORCH_AVAILABLE:
        advanced_tokenizer = Tokenizer(
            lowercase=True,
            remove_punctuation=True,
            remove_stopwords=True,
            language='english'
        )
        
        texts = [
            "The first example sentence for demonstration.",
            "The second example is a bit longer to show padding."
        ]
        
        # Build vocabulary
        advanced_tokenizer.build_vocab(texts, min_count=1)
        
        # Encode as tensors with padding
        encoded = advanced_tokenizer.batch_encode_as_tensors(
            texts, 
            padding=True,
            return_attention_mask=True
        )
        
        print(f"Input IDs shape: {encoded['input_ids'].shape}")
        print(f"Attention mask shape: {encoded['attention_mask'].shape}")
        
        # Decode back
        decoded = advanced_tokenizer.decode(encoded['input_ids'][0].tolist())
        print(f"Decoded text: {decoded}")
        
        # Example with Hugging Face datasets and PyTorch DataLoaders
        try:
            import datasets
            from torch.utils.data import Dataset, DataLoader
            
            # Load the rotten_tomatoes dataset
            print("Loading Rotten Tomatoes dataset...")
            dataset = datasets.load_dataset("rotten_tomatoes")
            
            # Create a simple dataset class
            class TextClassificationDataset(Dataset):
                def __init__(self, texts, labels, tokenizer, max_length=128):
                    self.texts = texts
                    self.labels = labels
                    self.tokenizer = tokenizer
                    self.max_length = max_length
                
                def __len__(self):
                    return len(self.texts)
                
                def __getitem__(self, idx):
                    text = self.texts[idx]
                    label = self.labels[idx]
                    
                    # Encode text
                    encoding = self.tokenizer.batch_encode_as_tensors(
                        [text], 
                        add_special_tokens=True,
                        padding=True,
                        truncation=True,
                        max_length=self.max_length,
                        return_attention_mask=True,
                        return_tensors=None  # We'll convert to tensors later
                    )
                    
                    # Convert to tensors
                    return {
                        'input_ids': torch.tensor(encoding['input_ids'][0]),
                        'attention_mask': torch.tensor(encoding['attention_mask'][0]),
                        'label': torch.tensor(label)
                    }
            
            # Build vocabulary from training data
            # Get text samples for building vocab
            train_texts = dataset['train']['text'][:1000]  # Using first 1000 samples for demonstration
            
            # Create tokenizer and build vocab
            rt_tokenizer = Tokenizer(
                lowercase=True,
                remove_punctuation=True,
                remove_stopwords=False
            )
            
            rt_tokenizer.build_vocab(train_texts, min_count=2)
            print(f"Built vocabulary of size: {len(rt_tokenizer.token_to_id)}")
            
            # Create PyTorch datasets
            train_dataset = TextClassificationDataset(
                dataset['train']['text'],
                dataset['train']['label'],
                rt_tokenizer
            )
            
            val_dataset = TextClassificationDataset(
                dataset['validation']['text'],
                dataset['validation']['label'],
                rt_tokenizer
            )
            
            # Create DataLoaders
            batch_size = 16
            
            train_loader = DataLoader(
                train_dataset,
                batch_size=batch_size,
                shuffle=True,
                collate_fn=lambda batch: Tokenizer.collate_fn(batch, rt_tokenizer.pad_token_id)
            )
            
            val_loader = DataLoader(
                val_dataset,
                batch_size=batch_size,
                shuffle=False,
                collate_fn=lambda batch: Tokenizer.collate_fn(batch, rt_tokenizer.pad_token_id)
            )
            
            # Show a batch
            print("\nExample batch from DataLoader:")
            for batch in train_loader:
                print(f"Input IDs shape: {batch['input_ids'].shape}")
                print(f"Attention mask shape: {batch['attention_mask'].shape}")
                print(f"Labels shape: {batch['label'].shape}")
                
                # Decode a sample
                sample_idx = 0
                print(f"\nSample text decoded: {rt_tokenizer.decode(batch['input_ids'][sample_idx].tolist(), skip_special_tokens=True)}")
                print(f"Sample label: {batch['label'][sample_idx].item()}")
                break
                
        except ImportError:
            print("Hugging Face datasets library not available. Install with 'pip install datasets'")


"""
Sequence-to-Sequence model with BiLSTM encoder and attention mechanism
for tasks like machine translation, text summarization, etc.
"""

import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import random
from typing import Tuple, List, Dict, Optional, Union

class Encoder(nn.Module):
    """Bidirectional LSTM encoder for sequence-to-sequence tasks."""
    
    def __init__(
        self, 
        input_dim: int, 
        emb_dim: int, 
        hidden_dim: int, 
        n_layers: int, 
        dropout: float
    ):
        """
        Initialize the encoder.
        
        Args:
            input_dim: Size of the input vocabulary
            emb_dim: Dimension of embeddings
            hidden_dim: Dimension of hidden states
            n_layers: Number of LSTM layers
            dropout: Dropout probability
        """
        super().__init__()
        
        self.embedding = nn.Embedding(input_dim, emb_dim)
        self.rnn = nn.LSTM(
            emb_dim, 
            hidden_dim, 
            num_layers=n_layers, 
            bidirectional=True, 
            dropout=dropout if n_layers > 1 else 0
        )
        self.dropout = nn.Dropout(dropout)
        self.hidden_dim = hidden_dim
        self.n_layers = n_layers
        
    def forward(
        self, 
        src: torch.Tensor, 
        src_lengths: Optional[torch.Tensor] = None
    ) -> Tuple[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """
        Forward pass of the encoder.
        
        Args:
            src: Source sequence [seq_len, batch_size]
            src_lengths: Length of each sequence in the batch (for packing)
            
        Returns:
            encoder_outputs: All hidden states [seq_len, batch_size, hidden_dim * 2]
            (hidden, cell): Final hidden and cell states 
                           [n_layers * 2, batch_size, hidden_dim]
        """
        # Embed tokens
        embedded = self.dropout(self.embedding(src))
        
        # Use packed sequences for variable length inputs
        if src_lengths is not None:
            packed_embedded = nn.utils.rnn.pack_padded_sequence(
                embedded, src_lengths.cpu(), enforce_sorted=False
            )
            packed_outputs, (hidden, cell) = self.rnn(packed_embedded)
            encoder_outputs, _ = nn.utils.rnn.pad_packed_sequence(packed_outputs)
        else:
            encoder_outputs, (hidden, cell) = self.rnn(embedded)
        
        # Combine the forward and backward final hidden states
        hidden_forward = hidden[::2]
        hidden_backward = hidden[1::2]
        hidden = torch.cat([hidden_forward, hidden_backward], dim=2)
        
        # Same for cell state
        cell_forward = cell[::2]
        cell_backward = cell[1::2]
        cell = torch.cat([cell_forward, cell_backward], dim=2)
        
        return encoder_outputs, (hidden, cell)


class Attention(nn.Module):
    """Attention mechanism for focusing on relevant parts of the input sequence."""
    
    def __init__(
        self, 
        enc_hidden_dim: int, 
        dec_hidden_dim: int,
        attn_dim: int
    ):
        """
        Initialize the attention mechanism.
        
        Args:
            enc_hidden_dim: Dimension of encoder hidden states
            dec_hidden_dim: Dimension of decoder hidden states
            attn_dim: Dimension of attention weights
        """
        super().__init__()
        
        self.attn = nn.Linear((enc_hidden_dim * 2) + dec_hidden_dim, attn_dim)
        self.v = nn.Linear(attn_dim, 1, bias=False)
        
    def forward(
        self, 
        hidden: torch.Tensor, 
        encoder_outputs: torch.Tensor, 
        mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Forward pass of the attention mechanism.
        
        Args:
            hidden: Current decoder hidden state [batch_size, dec_hidden_dim]
            encoder_outputs: All encoder hidden states [src_len, batch_size, enc_hidden_dim * 2]
            mask: Mask to avoid attending to padding tokens [batch_size, src_len]
            
        Returns:
            attention: Attention weights [batch_size, src_len]
        """
        src_len = encoder_outputs.shape[0]
        batch_size = encoder_outputs.shape[1]
        
        # Repeat the hidden state src_len times
        hidden = hidden.unsqueeze(1).repeat(1, src_len, 1)
        
        # Transpose encoder outputs to align with hidden
        encoder_outputs = encoder_outputs.permute(1, 0, 2)
        
        # Concatenate hidden state and encoder outputs
        energy = torch.tanh(
            self.attn(torch.cat((hidden, encoder_outputs), dim=2))
        )
        
        # Get attention score for each position
        attention = self.v(energy).squeeze(2)
        
        # Mask out padding positions
        if mask is not None:
            attention = attention.masked_fill(mask == 0, -1e10)
        
        # Softmax to get attention weights
        attention = F.softmax(attention, dim=1)
        
        return attention


class Decoder(nn.Module):
    """LSTM decoder with attention for sequence-to-sequence tasks."""
    
    def __init__(
        self, 
        output_dim: int, 
        emb_dim: int, 
        enc_hidden_dim: int,
        dec_hidden_dim: int, 
        attn_dim: int,
        n_layers: int, 
        dropout: float
    ):
        """
        Initialize the decoder.
        
        Args:
            output_dim: Size of the output vocabulary
            emb_dim: Dimension of embeddings
            enc_hidden_dim: Dimension of encoder hidden states
            dec_hidden_dim: Dimension of decoder hidden states
            attn_dim: Dimension of attention weights
            n_layers: Number of LSTM layers
            dropout: Dropout probability
        """
        super().__init__()
        
        self.output_dim = output_dim
        self.attention = Attention(enc_hidden_dim, dec_hidden_dim, attn_dim)
        
        self.embedding = nn.Embedding(output_dim, emb_dim)
        self.rnn = nn.LSTM(
            emb_dim + (enc_hidden_dim * 2), 
            dec_hidden_dim, 
            num_layers=n_layers, 
            dropout=dropout if n_layers > 1 else 0
        )
        self.fc_out = nn.Linear((enc_hidden_dim * 2) + dec_hidden_dim + emb_dim, output_dim)
        self.dropout = nn.Dropout(dropout)
        
    def forward(
        self, 
        input: torch.Tensor, 
        hidden: Tuple[torch.Tensor, torch.Tensor], 
        encoder_outputs: torch.Tensor,
        mask: Optional[torch.Tensor] = None
    ) -> Tuple[torch.Tensor, Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:
        """
        Forward pass of the decoder for a single time step.
        
        Args:
            input: Current input token [batch_size]
            hidden: Current hidden and cell states ([layers, batch_size, dec_hidden_dim], [...])
            encoder_outputs: All encoder hidden states [src_len, batch_size, enc_hidden_dim * 2]
            mask: Mask to avoid attending to padding tokens [batch_size, src_len]
            
        Returns:
            prediction: Prediction for current time step [batch_size, output_dim]
            hidden: Updated hidden and cell states
            attention: Attention weights [batch_size, src_len]
        """
        input = input.unsqueeze(0)
        
        # Embed token
        embedded = self.dropout(self.embedding(input))
        
        # Calculate attention weights
        attn_weights = self.attention(hidden[0][-1], encoder_outputs, mask)
        
        # Reshape attention weights and encoder outputs for bmm
        attn_weights = attn_weights.unsqueeze(1)
        encoder_outputs = encoder_outputs.permute(1, 0, 2)
        
        # Calculate context vector using attention weights
        context = torch.bmm(attn_weights, encoder_outputs)
        context = context.permute(1, 0, 2)
        
        # Concatenate embedding and context vector as input to RNN
        rnn_input = torch.cat((embedded, context), dim=2)
        
        # Pass through RNN
        output, hidden = self.rnn(rnn_input, hidden)
        
        # Reshape for concatenation
        output = output.squeeze(0)
        embedded = embedded.squeeze(0)
        context = context.squeeze(0)
        
        # Make prediction
        prediction = self.fc_out(torch.cat((output, context, embedded), dim=1))
        
        return prediction, hidden, attn_weights.squeeze(1)


class Seq2Seq(nn.Module):
    """Sequence-to-sequence model with attention for various NLP tasks."""
    
    def __init__(
        self, 
        encoder: Encoder, 
        decoder: Decoder, 
        device: torch.device,
        pad_idx: int,
        sos_idx: int,
        eos_idx: int
    ):
        """
        Initialize the sequence-to-sequence model.
        
        Args:
            encoder: Encoder module
            decoder: Decoder module
            device: Device to use (cuda/cpu)
            pad_idx: Padding token index
            sos_idx: Start of sequence token index
            eos_idx: End of sequence token index
        """
        super().__init__()
        
        self.encoder = encoder
        self.decoder = decoder
        self.device = device
        self.pad_idx = pad_idx
        self.sos_idx = sos_idx
        self.eos_idx = eos_idx
        
    def create_mask(self, src: torch.Tensor) -> torch.Tensor:
        """
        Create a mask for the source sequence padding tokens.
        
        Args:
            src: Source sequence [seq_len, batch_size]
            
        Returns:
            mask: Binary mask (1 for token, 0 for pad) [batch_size, seq_len]
        """
        # Create mask where 1 means token and 0 means padding
        mask = (src != self.pad_idx).permute(1, 0)
        return mask
        
    def forward(
        self, 
        src: torch.Tensor, 
        src_lengths: Optional[torch.Tensor], 
        trg: torch.Tensor, 
        teacher_forcing_ratio: float = 0.5
    ) -> torch.Tensor:
        """
        Forward pass of the sequence-to-sequence model.
        
        Args:
            src: Source sequence [seq_len, batch_size]
            src_lengths: Length of each source sequence (for packing)
            trg: Target sequence [seq_len, batch_size]
            teacher_forcing_ratio: Probability of using teacher forcing
            
        Returns:
            outputs: Predictions for each token in target sequence [seq_len, batch_size, output_dim]
        """
        batch_size = trg.shape[1]
        trg_len = trg.shape[0]
        trg_vocab_size = self.decoder.output_dim
        
        # Create tensor to store decoder outputs
        outputs = torch.zeros(trg_len, batch_size, trg_vocab_size).to(self.device)
        
        # Create source mask
        mask = self.create_mask(src)
        
        # Encode the source sequence
        encoder_outputs, hidden = self.encoder(src, src_lengths)
        
        # First input to the decoder is the <sos> token
        input = trg[0, :]
        
        # For each position in the target sequence
        for t in range(1, trg_len):
            # Pass through decoder
            output, hidden, _ = self.decoder(input, hidden, encoder_outputs, mask)
            
            # Store prediction
            outputs[t] = output
            
            # Determine next input: teacher forcing or highest probability
            teacher_force = random.random() < teacher_forcing_ratio
            
            # Get the highest predicted token
            top1 = output.argmax(1)
            
            # Use teacher forcing or own prediction
            input = trg[t] if teacher_force else top1
            
        return outputs
    
    def translate(
        self, 
        src: torch.Tensor, 
        src_lengths: Optional[torch.Tensor] = None, 
        max_len: int = 50
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Generate translations for the source sequence.
        
        Args:
            src: Source sequence [seq_len, batch_size]
            src_lengths: Length of each sequence (for packing)
            max_len: Maximum length of the generated sequence
            
        Returns:
            translations: Generated sequences [max_len, batch_size]
            attention: Attention weights for each position [max_len, batch_size, src_len]
        """
        batch_size = src.shape[1]
        
        # Create source mask
        mask = self.create_mask(src)
        
        # Encode the source sequence
        encoder_outputs, hidden = self.encoder(src, src_lengths)
        
        # First input to the decoder is the <sos> token
        input = torch.full((batch_size,), self.sos_idx, dtype=torch.long).to(self.device)
        
        # Tensors to store translation and attention
        translations = torch.zeros(max_len, batch_size, dtype=torch.long).to(self.device)
        translations[0] = input
        attention_weights = torch.zeros(max_len, batch_size, src.shape[0]).to(self.device)
        
        # For each position in the output sequence
        for t in range(1, max_len):
            # Pass through decoder
            output, hidden, attn = self.decoder(input, hidden, encoder_outputs, mask)
            
            # Store attention weights
            attention_weights[t] = attn
            
            # Get the most likely next token
            pred_token = output.argmax(1)
            
            # Store prediction
            translations[t] = pred_token
            
            # Check if all sequences have reached <eos>
            if (pred_token == self.eos_idx).all():
                break
                
            # Next input is the predicted token
            input = pred_token
            
        return translations, attention_weights


def train_seq2seq(
    model: Seq2Seq,
    iterator: torch.utils.data.DataLoader,
    optimizer: torch.optim.Optimizer,
    criterion: nn.Module,
    clip: float = 1.0,
    teacher_forcing_ratio: float = 0.5
) -> float:
    """
    Train the sequence-to-sequence model for one epoch.
    
    Args:
        model: Seq2Seq model
        iterator: DataLoader with training data
        optimizer: Optimizer for updating weights
        criterion: Loss function
        clip: Gradient clipping value
        teacher_forcing_ratio: Probability of using teacher forcing
        
    Returns:
        epoch_loss: Average loss for the epoch
    """
    model.train()
    epoch_loss = 0
    
    for batch in iterator:
        src, src_lengths = batch.src
        trg = batch.trg
        
        optimizer.zero_grad()
        
        # Forward pass
        output = model(src, src_lengths, trg, teacher_forcing_ratio)
        
        # Get loss, ignoring the <sos> token
        output_dim = output.shape[-1]
        output = output[1:].view(-1, output_dim)
        trg = trg[1:].view(-1)
        
        loss = criterion(output, trg)
        
        # Backward pass
        loss.backward()
        
        # Clip gradients
        torch.nn.utils.clip_grad_norm_(model.parameters(), clip)
        
        # Update weights
        optimizer.step()
        
        epoch_loss += loss.item()
        
    return epoch_loss / len(iterator)


def evaluate_seq2seq(
    model: Seq2Seq,
    iterator: torch.utils.data.DataLoader,
    criterion: nn.Module
) -> float:
    """
    Evaluate the sequence-to-sequence model.
    
    Args:
        model: Seq2Seq model
        iterator: DataLoader with evaluation data
        criterion: Loss function
        
    Returns:
        epoch_loss: Average loss for the evaluation
    """
    model.eval()
    epoch_loss = 0
    
    with torch.no_grad():
        for batch in iterator:
            src, src_lengths = batch.src
            trg = batch.trg
            
            # Forward pass with 0 teacher forcing
            output = model(src, src_lengths, trg, 0)
            
            # Get loss, ignoring the <sos> token
            output_dim = output.shape[-1]
            output = output[1:].view(-1, output_dim)
            trg = trg[1:].view(-1)
            
            loss = criterion(output, trg)
            epoch_loss += loss.item()
    
    return epoch_loss / len(iterator)


# Example usage
def example_usage():
    """Example of how to use the Seq2Seq model for machine translation without torchtext."""
    import torch
    from torch.utils.data import Dataset, DataLoader
    import spacy
    import pandas as pd
    import numpy as np
    from collections import Counter
    
    # Set up spaCy tokenizers
    try:
        spacy_de = spacy.load('de_core_news_sm')
        spacy_en = spacy.load('en_core_web_sm')
    except OSError:
        print("Spacy models need to be downloaded first via:")
        print("python -m spacy download de_core_news_sm")
        print("python -m spacy download en_core_web_sm")
        return
    
    def tokenize_de(text):
        return [tok.text for tok in spacy_de.tokenizer(text)]
        
    def tokenize_en(text):
        return [tok.text for tok in spacy_en.tokenizer(text)]
        
    class Vocabulary:
        """Simple vocabulary class to replace torchtext.Field functionality."""
        def __init__(self, freq_threshold=2):
            self.itos = {0: '<pad>', 1: '<sos>', 2: '<eos>', 3: '<unk>'}
            self.stoi = {'<pad>': 0, '<sos>': 1, '<eos>': 2, '<unk>': 3}
            self.freq_threshold = freq_threshold
            
        def __len__(self):
            return len(self.itos)
            
        def build_vocabulary(self, sentence_list):
            frequencies = Counter()
            idx = 4
            
            for sentence in sentence_list:
                for word in sentence:
                    frequencies[word] += 1
                    
            for word, count in frequencies.items():
                if count >= self.freq_threshold:
                    self.stoi[word] = idx
                    self.itos[idx] = word
                    idx += 1
    
    class TranslationDataset(Dataset):
        """Custom dataset for translation data."""
        def __init__(self, src_texts, trg_texts, src_vocab, trg_vocab):
            self.src_texts = src_texts
            self.trg_texts = trg_texts
            self.src_vocab = src_vocab
            self.trg_vocab = trg_vocab
            
        def __len__(self):
            return len(self.src_texts)
            
        def __getitem__(self, index):
            src_text = self.src_texts[index]
            trg_text = self.trg_texts[index]
            
            # Convert tokens to indices
            src_indices = [self.src_vocab.stoi.get(token, self.src_vocab.stoi['<unk>']) 
                           for token in src_text]
            trg_indices = [self.trg_vocab.stoi.get(token, self.trg_vocab.stoi['<unk>']) 
                           for token in trg_text]
            
            # Add <sos> and <eos> tokens
            src_indices = [self.src_vocab.stoi['<sos>']] + src_indices + [self.src_vocab.stoi['<eos>']]
            trg_indices = [self.trg_vocab.stoi['<sos>']] + trg_indices + [self.trg_vocab.stoi['<eos>']]
            
            return {
                'src': src_indices,
                'trg': trg_indices,
                'src_len': len(src_indices)
            }
    
    def collate_fn(batch):
        """Custom collate function for variable length sequences."""
        # Sort batch by source sequence length in descending order
        batch = sorted(batch, key=lambda x: x['src_len'], reverse=True)
        
        # Get lengths
        src_lengths = torch.tensor([item['src_len'] for item in batch])
        
        # Pad sequences
        max_src_len = max(item['src_len'] for item in batch)
        max_trg_len = max(len(item['trg']) for item in batch)
        
        # Create padded tensors
        padded_src = torch.full((max_src_len, len(batch)), src_vocab.stoi['<pad>'], dtype=torch.long)
        padded_trg = torch.full((max_trg_len, len(batch)), trg_vocab.stoi['<pad>'], dtype=torch.long)
        
        # Fill in the data
        for i, item in enumerate(batch):
            src = torch.tensor(item['src'])
            trg = torch.tensor(item['trg'])
            
            padded_src[:src.size(0), i] = src
            padded_trg[:trg.size(0), i] = trg
            
        return padded_src, src_lengths, padded_trg
    
    # Load and preprocess data
    # Replace with your actual data loading code
    try:
        # This is a simplified example - replace with your actual data
        df_train = pd.read_csv('path/to/data/train.csv')
        df_valid = pd.read_csv('path/to/data/valid.csv')
        df_test = pd.read_csv('path/to/data/test.csv')
    except FileNotFoundError:
        print("Example data files not found. Creating dummy data for demonstration.")
        # Create dummy data for demonstration
        src_examples = [
            "Ein kleines Mädchen steht vor einem Gebäude.",
            "Der Mann geht zur Arbeit.",
            "Ich liebe Programmierung."
        ]
        trg_examples = [
            "A little girl stands in front of a building.",
            "The man goes to work.",
            "I love programming."
        ]
        
        # Create dataframes with dummy data
        df_train = pd.DataFrame({'src': src_examples, 'trg': trg_examples})
        df_valid = df_train.copy()  # Use same data for demo
        df_test = df_train.copy()  # Use same data for demo
    
    # Tokenize data
    train_src_tokenized = [tokenize_de(text) for text in df_train['src']]
    train_trg_tokenized = [tokenize_en(text) for text in df_train['trg']]
    valid_src_tokenized = [tokenize_de(text) for text in df_valid['src']]
    valid_trg_tokenized = [tokenize_en(text) for text in df_valid['trg']]
    test_src_tokenized = [tokenize_de(text) for text in df_test['src']]
    test_trg_tokenized = [tokenize_en(text) for text in df_test['trg']]
    
    # Build vocabularies
    src_vocab = Vocabulary()
    trg_vocab = Vocabulary()
    src_vocab.build_vocabulary(train_src_tokenized)
    trg_vocab.build_vocabulary(train_trg_tokenized)
    
    # Create datasets
    train_dataset = TranslationDataset(train_src_tokenized, train_trg_tokenized, src_vocab, trg_vocab)
    valid_dataset = TranslationDataset(valid_src_tokenized, valid_trg_tokenized, src_vocab, trg_vocab)
    test_dataset = TranslationDataset(test_src_tokenized, test_trg_tokenized, src_vocab, trg_vocab)
    
    # Create data loaders
    BATCH_SIZE = 64
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    
    train_iterator = DataLoader(
        train_dataset, 
        batch_size=BATCH_SIZE, 
        collate_fn=collate_fn, 
        shuffle=True
    )
    valid_iterator = DataLoader(
        valid_dataset, 
        batch_size=BATCH_SIZE, 
        collate_fn=collate_fn
    )
    test_iterator = DataLoader(
        test_dataset, 
        batch_size=BATCH_SIZE, 
        collate_fn=collate_fn
    )
    
    # Define model hyperparameters
    INPUT_DIM = len(src_vocab)
    OUTPUT_DIM = len(trg_vocab)
    ENC_EMB_DIM = 256
    DEC_EMB_DIM = 256
    ENC_HIDDEN_DIM = 512
    DEC_HIDDEN_DIM = 512
    ATTN_DIM = 64
    ENC_LAYERS = 2
    DEC_LAYERS = 2
    ENC_DROPOUT = 0.5
    DEC_DROPOUT = 0.5
    PAD_IDX = src_vocab.stoi['<pad>']
    SOS_IDX = trg_vocab.stoi['<sos>']
    EOS_IDX = trg_vocab.stoi['<eos>']
    
    # Create model components
    encoder = Encoder(INPUT_DIM, ENC_EMB_DIM, ENC_HIDDEN_DIM, ENC_LAYERS, ENC_DROPOUT)
    decoder = Decoder(OUTPUT_DIM, DEC_EMB_DIM, ENC_HIDDEN_DIM, DEC_HIDDEN_DIM, ATTN_DIM, DEC_LAYERS, DEC_DROPOUT)
    model = Seq2Seq(encoder, decoder, device, PAD_IDX, SOS_IDX, EOS_IDX).to(device)
    
    # Initialize weights
    def init_weights(m):
        for name, param in m.named_parameters():
            if 'weight' in name:
                nn.init.normal_(param.data, mean=0, std=0.01)
            else:
                nn.init.constant_(param.data, 0)
    
    model.apply(init_weights)
    
    # Define optimizer and loss
    optimizer = optim.Adam(model.parameters())
    criterion = nn.CrossEntropyLoss(ignore_index=PAD_IDX)
    
    # Train model
    N_EPOCHS = 10
    CLIP = 1.0
    
    best_valid_loss = float('inf')
    
    for epoch in range(N_EPOCHS):
        train_loss = train_seq2seq(model, train_iterator, optimizer, criterion, CLIP)
        valid_loss = evaluate_seq2seq(model, valid_iterator, criterion)
        
        if valid_loss < best_valid_loss:
            best_valid_loss = valid_loss
            torch.save(model.state_dict(), 'seq2seq-model.pt')
        
        print(f'Epoch: {epoch+1}')
        print(f'\tTrain Loss: {train_loss:.3f}')
        print(f'\tValid Loss: {valid_loss:.3f}')
    
    # Example function to translate a sentence
    def translate_sentence(sentence, src_vocab, trg_vocab, model, device, max_len=50):
        model.eval()
        
        # Tokenize the source sentence if it's a string
        if isinstance(sentence, str):
            tokens = tokenize_de(sentence)
        else:
            tokens = sentence
            
        # Convert to indices
        src_indices = [src_vocab.stoi.get(token, src_vocab.stoi['<unk>']) for token in tokens]
        
        # Add <sos> and <eos> tokens
        src_indices = [src_vocab.stoi['<sos>']] + src_indices + [src_vocab.stoi['<eos>']]
        
        # Convert to tensor
        src_tensor = torch.LongTensor(src_indices).unsqueeze(1).to(device)
        src_lengths = torch.LongTensor([len(src_indices)]).to(device)
        
        # Generate translation
        with torch.no_grad():
            translations, attention = model.translate(src_tensor, src_lengths, max_len)
        
        # Convert indices to tokens
        trg_tokens = [trg_vocab.itos[idx] for idx in translations[:, 0]]
        
        # Remove <sos> and find position of <eos>
        try:
            eos_pos = trg_tokens.index(trg_vocab.itos[EOS_IDX])
            trg_tokens = trg_tokens[1:eos_pos]
        except ValueError:
            trg_tokens = trg_tokens[1:]
            
        return trg_tokens, attention
    
    # Test with an example sentence
    test_sentence = "Ein kleines Mädchen steht vor einem Gebäude."
    translation, attention = translate_sentence(test_sentence, src_vocab, trg_vocab, model, device)
    print(f'Source: {test_sentence}')
    print(f'Translation: {" ".join(translation)}')


if __name__ == "__main__":
    # Uncomment to run the example
    # example_usage()
    pass



"""
Inception-ResNet implementations based on "Inception-v4, Inception-ResNet and the Impact of Residual Connections on Learning"
https://arxiv.org/abs/1602.07261
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Optional, Tuple, List, Callable


class BasicConv2d(nn.Module):
    """Basic convolution module: Conv2d + BN + ReLU"""
    def __init__(
        self, 
        in_channels: int, 
        out_channels: int, 
        **kwargs
    ) -> None:
        super().__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, bias=False, **kwargs)
        self.bn = nn.BatchNorm2d(out_channels, eps=0.001)
        self.relu = nn.ReLU(inplace=True)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class Stem(nn.Module):
    """Stem for both Inception-ResNet v1 and v2"""
    def __init__(self, in_channels: int = 3) -> None:
        super().__init__()
        self.conv2d_1a = BasicConv2d(in_channels, 32, kernel_size=3, stride=2, padding=0)
        self.conv2d_2a = BasicConv2d(32, 32, kernel_size=3, stride=1, padding=0)
        self.conv2d_2b = BasicConv2d(32, 64, kernel_size=3, stride=1, padding=1)
        self.maxpool_3a = nn.MaxPool2d(kernel_size=3, stride=2, padding=0)
        self.conv2d_3b = BasicConv2d(64, 80, kernel_size=1, stride=1, padding=0)
        self.conv2d_4a = BasicConv2d(80, 192, kernel_size=3, stride=1, padding=0)
        self.conv2d_4b = BasicConv2d(192, 256, kernel_size=3, stride=2, padding=0)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv2d_1a(x)
        x = self.conv2d_2a(x)
        x = self.conv2d_2b(x)
        x = self.maxpool_3a(x)
        x = self.conv2d_3b(x)
        x = self.conv2d_4a(x)
        x = self.conv2d_4b(x)
        return x


class InceptionResnetA(nn.Module):
    """Inception-ResNet-A block for Inception-ResNet v1 and v2"""
    def __init__(self, in_channels: int, scale: float = 1.0) -> None:
        super().__init__()
        self.scale = scale
        
        self.branch0 = BasicConv2d(in_channels, 32, kernel_size=1, stride=1, padding=0)
        
        self.branch1_0 = BasicConv2d(in_channels, 32, kernel_size=1, stride=1, padding=0)
        self.branch1_1 = BasicConv2d(32, 32, kernel_size=3, stride=1, padding=1)
        
        self.branch2_0 = BasicConv2d(in_channels, 32, kernel_size=1, stride=1, padding=0)
        self.branch2_1 = BasicConv2d(32, 48, kernel_size=3, stride=1, padding=1)
        self.branch2_2 = BasicConv2d(48, 64, kernel_size=3, stride=1, padding=1)
        
        self.conv = nn.Conv2d(128, in_channels, kernel_size=1, stride=1, padding=0, bias=False)
        self.relu = nn.ReLU(inplace=True)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x0 = self.branch0(x)
        
        x1 = self.branch1_0(x)
        x1 = self.branch1_1(x1)
        
        x2 = self.branch2_0(x)
        x2 = self.branch2_1(x2)
        x2 = self.branch2_2(x2)
        
        mixed = torch.cat([x0, x1, x2], dim=1)
        up = self.conv(mixed)
        
        # Apply scaling to stabilize training
        x += self.scale * up
        x = self.relu(x)
        return x


class InceptionResnetB(nn.Module):
    """Inception-ResNet-B block for Inception-ResNet v1 and v2"""
    def __init__(self, in_channels: int, scale: float = 1.0) -> None:
        super().__init__()
        self.scale = scale
        
        self.branch0 = BasicConv2d(in_channels, 192, kernel_size=1, stride=1, padding=0)
        
        self.branch1_0 = BasicConv2d(in_channels, 128, kernel_size=1, stride=1, padding=0)
        self.branch1_1 = BasicConv2d(128, 160, kernel_size=(1, 7), stride=1, padding=(0, 3))
        self.branch1_2 = BasicConv2d(160, 192, kernel_size=(7, 1), stride=1, padding=(3, 0))
        
        self.conv = nn.Conv2d(384, in_channels, kernel_size=1, stride=1, padding=0, bias=False)
        self.relu = nn.ReLU(inplace=True)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x0 = self.branch0(x)
        
        x1 = self.branch1_0(x)
        x1 = self.branch1_1(x1)
        x1 = self.branch1_2(x1)
        
        mixed = torch.cat([x0, x1], dim=1)
        up = self.conv(mixed)
        
        # Apply scaling to stabilize training
        x += self.scale * up
        x = self.relu(x)
        return x


class InceptionResnetC(nn.Module):
    """Inception-ResNet-C block for Inception-ResNet v1 and v2"""
    def __init__(self, in_channels: int, scale: float = 1.0, activation: bool = True) -> None:
        super().__init__()
        self.scale = scale
        self.activation = activation
        
        self.branch0 = BasicConv2d(in_channels, 192, kernel_size=1, stride=1, padding=0)
        
        self.branch1_0 = BasicConv2d(in_channels, 192, kernel_size=1, stride=1, padding=0)
        self.branch1_1 = BasicConv2d(192, 224, kernel_size=(1, 3), stride=1, padding=(0, 1))
        self.branch1_2 = BasicConv2d(224, 256, kernel_size=(3, 1), stride=1, padding=(1, 0))
        
        self.conv = nn.Conv2d(448, in_channels, kernel_size=1, stride=1, padding=0, bias=False)
        if self.activation:
            self.relu = nn.ReLU(inplace=True)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x0 = self.branch0(x)
        
        x1 = self.branch1_0(x)
        x1 = self.branch1_1(x1)
        x1 = self.branch1_2(x1)
        
        mixed = torch.cat([x0, x1], dim=1)
        up = self.conv(mixed)
        
        # Apply scaling to stabilize training
        x += self.scale * up
        if self.activation:
            x = self.relu(x)
        return x


class ReductionA(nn.Module):
    """Reduction-A block for Inception-ResNet v1 and v2"""
    def __init__(self, in_channels: int, k: int, l: int, m: int, n: int) -> None:
        super().__init__()
        self.branch0 = BasicConv2d(in_channels, n, kernel_size=3, stride=2, padding=0)
        
        self.branch1_0 = BasicConv2d(in_channels, k, kernel_size=1, stride=1, padding=0)
        self.branch1_1 = BasicConv2d(k, l, kernel_size=3, stride=1, padding=1)
        self.branch1_2 = BasicConv2d(l, m, kernel_size=3, stride=2, padding=0)
        
        self.branch2 = nn.MaxPool2d(kernel_size=3, stride=2, padding=0)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x0 = self.branch0(x)
        x1 = self.branch1_0(x)
        x1 = self.branch1_1(x1)
        x1 = self.branch1_2(x1)
        x2 = self.branch2(x)
        return torch.cat([x0, x1, x2], dim=1)


class ReductionB(nn.Module):
    """Reduction-B block for Inception-ResNet v1 and v2"""
    def __init__(self, in_channels: int) -> None:
        super().__init__()
        self.branch0_0 = BasicConv2d(in_channels, 256, kernel_size=1, stride=1, padding=0)
        self.branch0_1 = BasicConv2d(256, 384, kernel_size=3, stride=2, padding=0)
        
        self.branch1_0 = BasicConv2d(in_channels, 256, kernel_size=1, stride=1, padding=0)
        self.branch1_1 = BasicConv2d(256, 288, kernel_size=3, stride=2, padding=0)
        
        self.branch2_0 = BasicConv2d(in_channels, 256, kernel_size=1, stride=1, padding=0)
        self.branch2_1 = BasicConv2d(256, 288, kernel_size=3, stride=1, padding=1)
        self.branch2_2 = BasicConv2d(288, 320, kernel_size=3, stride=2, padding=0)
        
        self.branch3 = nn.MaxPool2d(kernel_size=3, stride=2, padding=0)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x0 = self.branch0_0(x)
        x0 = self.branch0_1(x0)
        
        x1 = self.branch1_0(x)
        x1 = self.branch1_1(x1)
        
        x2 = self.branch2_0(x)
        x2 = self.branch2_1(x2)
        x2 = self.branch2_2(x2)
        
        x3 = self.branch3(x)
        
        return torch.cat([x0, x1, x2, x3], dim=1)


class InceptionResNetV1(nn.Module):
    """Inception-ResNet-v1 model"""
    def __init__(self, num_classes: int = 1000, dropout_prob: float = 0.2) -> None:
        super().__init__()
        # Stem
        self.stem = Stem()
        
        # Inception-ResNet-A
        self.mixed_5a = nn.Sequential(*[InceptionResnetA(256, scale=0.17) for _ in range(5)])
        
        # Reduction-A
        self.mixed_6a = ReductionA(256, 192, 192, 256, 384)
        
        # Inception-ResNet-B
        self.mixed_6b = nn.Sequential(*[InceptionResnetB(896, scale=0.10) for _ in range(10)])
        
        # Reduction-B
        self.mixed_7a = ReductionB(896)
        
        # Inception-ResNet-C
        self.mixed_8a = nn.Sequential(*[InceptionResnetC(1792, scale=0.20) for _ in range(5)])
        
        # Classification block
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout(dropout_prob)
        self.fc = nn.Linear(1792, num_classes)
        
        # Initialize weights
        for m in self.modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Stem
        x = self.stem(x)
        
        # Inception-ResNet-A and Reduction-A
        x = self.mixed_5a(x)
        x = self.mixed_6a(x)
        
        # Inception-ResNet-B and Reduction-B
        x = self.mixed_6b(x)
        x = self.mixed_7a(x)
        
        # Inception-ResNet-C
        x = self.mixed_8a(x)
        
        # Classification block
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.dropout(x)
        x = self.fc(x)
        
        return x


class InceptionResNetV2(nn.Module):
    """Inception-ResNet-v2 model"""
    def __init__(self, num_classes: int = 1000, dropout_prob: float = 0.2) -> None:
        super().__init__()
        # Stem
        self.stem = Stem()
        
        # Inception-ResNet-A
        self.mixed_5a = nn.Sequential(*[InceptionResnetA(256, scale=0.17) for _ in range(10)])
        
        # Reduction-A
        self.mixed_6a = ReductionA(256, 256, 256, 384, 384)
        
        # Inception-ResNet-B
        self.mixed_6b = nn.Sequential(*[InceptionResnetB(896, scale=0.10) for _ in range(20)])
        
        # Reduction-B
        self.mixed_7a = ReductionB(896)
        
        # Inception-ResNet-C
        self.mixed_8a = nn.Sequential(*[
            InceptionResnetC(1792, scale=0.20, activation=True) for _ in range(9)
        ])
        # Last block doesn't have activation
        self.mixed_8b = InceptionResnetC(1792, scale=0.20, activation=False)
        
        # Classification block
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout(dropout_prob)
        self.fc = nn.Linear(1792, num_classes)
        
        # Initialize weights
        for m in self.modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Stem
        x = self.stem(x)
        
        # Inception-ResNet-A and Reduction-A
        x = self.mixed_5a(x)
        x = self.mixed_6a(x)
        
        # Inception-ResNet-B and Reduction-B
        x = self.mixed_6b(x)
        x = self.mixed_7a(x)
        
        # Inception-ResNet-C
        x = self.mixed_8a(x)
        x = self.mixed_8b(x)
        
        # Classification block
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.dropout(x)
        x = self.fc(x)
        
        return x


def inception_resnet_v1(num_classes: int = 1000, pretrained: bool = False) -> InceptionResNetV1:
    """Inception-ResNet-v1 model"""
    model = InceptionResNetV1(num_classes=num_classes)
    if pretrained:
        # Pretrained weights are not officially available
        pass
    return model


def inception_resnet_v2(num_classes: int = 1000, pretrained: bool = False) -> InceptionResNetV2:
    """Inception-ResNet-v2 model"""
    model = InceptionResNetV2(num_classes=num_classes)
    if pretrained:
        # Pretrained weights are not officially available
        pass
    return model



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torchvision
import torchvision.transforms.v2 as transforms
from tqdm import tqdm

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
print(device)

transform = transforms.Compose([
    transforms.ToImage(), 
    transforms.ToDtype(torch.float32, scale=True),
    transforms.Resize((224,224), interpolation=transforms.InterpolationMode.BICUBIC),
    transforms.RandomHorizontalFlip(p=0.5),
    transforms.RandomRotation(30),
    transforms.AutoAugment(),
    transforms.Normalize((0.5,0.5,0.5), (0.5,0.5,0.5))
    ])

batch_size = 512
train_dataset = torchvision.datasets.CIFAR100(root="./data", train=True, download=True, transform=transform)
test_dataset = torchvision.datasets.CIFAR100(root="./data", train=False, download=True, transform=transform)
train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=batch_size, shuffle=True, num_workers=2)
test_loader = torch.utils.data.DataLoader(test_dataset, batch_size=batch_size, shuffle=True, num_workers=2)

class Block(nn.Module):
    expansion = 1
    def __init__(self, in_channels, out_channels, stride=1, downsample=None):
        super().__init__()
        self.relu = nn.ReLU(inplace=True)
        
        self.bn1 = nn.BatchNorm2d(in_channels)
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False)
        
        self.bn2 = nn.BatchNorm2d(out_channels)
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1, bias=False)

        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        skip = x
        
        out = self.bn1(x)
        out = self.relu(out)

        if self.downsample is not None: skip = self.downsample(out)

        out = self.conv1(out)
        
        out = self.bn2(out)
        out = self.relu(out)
        out = self.conv2(out)

        out += skip
        return out

class BottleNeck(nn.Module):
    expansion = 4
    def __init__(self, in_channels, out_channels, stride=1, downsample=None):
        super().__init__()
        width = out_channels
        self.relu = nn.ReLU(inplace=True)

        self.bn1 = nn.BatchNorm2d(in_channels)
        self.conv1 = nn.Conv2d(in_channels, width, kernel_size=1, bias=False)

        self.bn2 = nn.BatchNorm2d(width)
        self.conv2 = nn.Conv2d(width, width, kernel_size=3, stride=stride, padding=1, bias=False)

        self.bn3 = nn.BatchNorm2d(width)
        self.conv3 = nn.Conv2d(width, out_channels * self.expansion, kernel_size=3, padding=1, bias=False)

        self.downsample = downsample
        self.stride = stride
    
    def forward(self, x):
        skip = x
        
        out = self.bn1(x)
        out = self.relu(out)

        if self.downsample is not None: skip = self.downsample(out)
        
        out = self.conv1(out)

        out = self.bn2(out)
        out = self.relu(out)
        out = self.conv2(out)

        out = self.bn3(out)
        out = self.relu(out)
        out = self.conv3(out)

        out += skip
        return out
        

class ResNet(nn.Module):
    def __init__(self, block, num_classes, layers):
        super().__init__()
        self.width = 64
        self.conv1 = nn.Conv2d(3, self.width, kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = nn.BatchNorm2d(self.width)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(2)

        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2)

        self.bn_final = nn.BatchNorm2d(512 * block.expansion)
        self.relu_final = nn.ReLU(inplace=True)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)
        
    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        
        x = self.bn_final(x)
        x = self.relu_final(x)
        
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)
        
        return x
        
    def _make_layer(self, block, out_channels, blocks, stride=1):
        downsample = None
        if stride != 1 or self.width != out_channels * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(
                    self.width, 
                    out_channels * block.expansion, 
                    kernel_size=1, 
                    stride=stride, 
                    bias=False
                )
            )

        layers = []
        layers.append(block(self.width, out_channels, stride, downsample))
        self.width = out_channels * block.expansion
        
        for _ in range(1, blocks):
            layers.append(block(self.width, out_channels))

        return nn.Sequential(*layers)


model = ResNet(Block, 100, [2,2,2,2]).to(device)
load = False
if load:
    state_dict = torch.load('models/cifar100model1.pth', map_location='cuda')
    model.load_state_dict(state_dict)

criterion = nn.CrossEntropyLoss(label_smoothing=0.1)
optimizer = optim.SGD(model.parameters(), lr=0.1, momentum=0.9, weight_decay=5e-4)
# optimizer = optim.AdamW(model.parameters(), lr=0.01)
# scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=5, gamma=0.5)
def lr_lambda(epoch):
    if epoch < 20:  
        return 0.1
    elif epoch < 40:
        return 0.01
    elif epoch < 60:  
        return 0.001
    else:
        return 3e-4
scheduler = optim.lr_scheduler.LambdaLR(optimizer, lr_lambda)

EPOCH = 40
for epoch in range(EPOCH):
    print(f"\nEpoch {epoch + 1}/{EPOCH}")
    model.train()  # set model to training mode
    running_loss, correct, total = 0.0, 0, 0

    for i, (inputs, labels) in enumerate(tqdm(train_loader, total=len(train_loader))):
        inputs, labels = inputs.to(device), labels.to(device)
        optimizer.zero_grad()

        outputs = model(inputs)
        loss = criterion(outputs, labels)
        loss.backward()
        optimizer.step()

        running_loss += loss.item()
        _, predicted = torch.max(outputs, 1)
        total += labels.size(0)
        correct += (predicted == labels).sum().item()

    train_loss = running_loss / len(train_loader)
    train_acc = 100 * correct / total

    model.eval() 
    val_loss, val_correct, val_total = 0.0, 0, 0

    with torch.no_grad():  # disable gradient computation
        for inputs, labels in test_loader:
            inputs, labels = inputs.to(device), labels.to(device)

            outputs = model(inputs)
            loss = criterion(outputs, labels)
            val_loss += loss.item()

            _, predicted = torch.max(outputs, 1)
            val_total += labels.size(0)
            val_correct += (predicted == labels).sum().item()

    val_loss /= len(test_loader)
    val_acc = 100 * val_correct / val_total

    print(f"Train Loss: {train_loss:.4f} | Train Acc: {train_acc:.2f}%")
    print(f"Val   Loss: {val_loss:.4f} | Val   Acc: {val_acc:.2f}%")

    # scheduler.step()

PATH = 'models/cifar100model1.pth'
torch.save(model.state_dict(), PATH)

correct = 0
total = 0
with torch.no_grad():
    for data in test_loader:
        images, labels = data[0].to(device), data[1].to(device)
        outputs = model(images)
        _, predicted = torch.max(outputs, 1)
        total += labels.size(0)
        correct += (predicted == labels).sum().item()

print(f'Accuracy: {100 * correct // total} %')




from transformers import AutoTokenizer, AutoModelForCausalLM, AutoModelForSeq2SeqLM, pipeline
from tqdm import tqdm
from typing import Dict, Any, Optional, List, Union
import torch

# Dictionary of smaller LLMs with parameter counts (in billions), all under 5B
MODEL_PARAMS = {
    # Phi models (Microsoft)
    "microsoft/phi-1": 1.3,
    "microsoft/phi-1.5": 1.3,
    "microsoft/phi-2": 2.7,
    "microsoft/Phi-3-mini-4k-instruct": 3.8,
    "microsoft/Phi-3-mini-128k-instruct": 3.8,
    
    # Gemma models (Google)
    "google/gemma-2b": 2.0,
    "google/gemma-2b-it": 2.0,
    
    # TinyLlama models
    "TinyLlama/TinyLlama-1.1B-Chat-v1.0": 1.1,
    "TinyLlama/TinyLlama-1.1B-intermediate-step-1431k-3T": 1.1,
    
    # Smaller Mistral models
    "mistralai/Mistral-7B-v0.1": 4.0,  # Quantized version
    
    # FLAN models
    "google/flan-t5-small": 0.08,
    "google/flan-t5-base": 0.25,
    "google/flan-t5-large": 0.8,
    
    # GPT-2 models (OpenAI)
    "gpt2": 0.124,
    "gpt2-medium": 0.355,
    "gpt2-large": 0.774,
    "gpt2-xl": 1.5,
    
    # GPT-Neo models (EleutherAI)
    "EleutherAI/gpt-neo-125m": 0.125,
    "EleutherAI/gpt-neo-1.3B": 1.3,
    "EleutherAI/gpt-neo-2.7B": 2.7,
    
    # Google T5 models
    "google-t5/t5-small": 0.06,
    "google-t5/t5-base": 0.22,
    "google-t5/t5-large": 0.77,
    "google-t5/t5-3b": 3.0,
    "google/t5-v1_1-small": 0.06,
    "google/t5-v1_1-base": 0.22,
    "google/t5-v1_1-large": 0.77,
    "google/t5-v1_1-xl": 3.0,
    
    # Additional BLOOM models
    "bigscience/bloom-560m": 0.56,
    "bigscience/bloom-1b1": 1.1,
    "bigscience/bloom-1b7": 1.7,
    "bigscience/bloomz-560m": 0.56,
    "bigscience/bloomz-1b1": 1.1,
    "bigscience/bloomz-1b7": 1.7,
    
    # Smaller OPT models
    "facebook/opt-125m": 0.125,
    "facebook/opt-350m": 0.35,
    "facebook/opt-1.3b": 1.3,
    "facebook/opt-2.7b": 2.7,
}

# Dictionary mapping models to their appropriate pipeline type
MODEL_PIPELINES = {
    # Models that use text-generation pipeline (decoder-only models)
    "microsoft/phi": "text-generation",
    "google/gemma": "text-generation",
    "TinyLlama": "text-generation",
    "mistralai": "text-generation",
    "gpt2": "text-generation",
    "EleutherAI/gpt-neo": "text-generation",
    "bigscience/bloom": "text-generation",
    "bigscience/bloomz": "text-generation",
    "facebook/opt": "text-generation",
    
    # Models that use text2text-generation pipeline (encoder-decoder models)
    "t5": "text2text-generation",
    "google/flan-t5": "text2text-generation",
    "google-t5": "text2text-generation",
    "google/t5": "text2text-generation",
}

def get_pipeline_type(model_name: str) -> str:
    """
    Determine the appropriate pipeline type for a given model.
    
    Args:
        model_name: The HuggingFace model identifier
        
    Returns:
        Pipeline type: either "text-generation" or "text2text-generation"
    """
    for prefix, pipeline_type in MODEL_PIPELINES.items():
        if prefix.lower() in model_name.lower():
            return pipeline_type
    
    # Default to text-generation if no match is found
    return "text-generation"

def get_model_info(model_name: Optional[str] = None) -> Union[Dict[str, float], float]:
    """
    Get parameter count information for LLMs.
    
    Args:
        model_name: Specific model to get information about. If None, returns all models.
        
    Returns:
        Dictionary of all models with parameter counts, or a single count for the specified model.
    """
    if model_name is None:
        return MODEL_PARAMS
    
    if model_name in MODEL_PARAMS:
        return MODEL_PARAMS[model_name]
    
    return f"Model {model_name} not found in the parameter dictionary."

def eval_llm(
    data: Any,
    text_column: str = 'text',
    label_column: str = 'label',
    system_prompt: str = "You are a classifier that outputs exactly one word or phrase from a fixed set of options.",
    task_prompt: str = "Classify the following text into exactly one of these categories: {labels}. Text: ",
    model_name: str = "microsoft/Phi-3-mini-4k-instruct",
    stride: int = 1,
    device: Optional[str] = None,
    max_new_tokens: int = 10,
    labels: Optional[List[str]] = None
) -> Dict[str, Any]:
    """
    Evaluate text classification using zero-shot prompting with an LLM.
    
    Args:
        data: DataFrame or dictionary containing text data and labels
        text_column: Name of the column containing text to classify
        label_column: Name of the column containing ground truth labels
        system_prompt: System prompt for the model
        task_prompt: Prompt template for the classification task (use {labels} to insert options)
        model_name: HuggingFace model identifier
        stride: Process every nth sample
        device: Computing device (auto-detected if None)
        max_new_tokens: Maximum new tokens to generate
        labels: List of possible label values (if None, will be extracted from data)
        
    Returns:
        Dictionary with evaluation metrics
    """
    # Determine device
    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    
    # Extract unique labels if not provided
    if labels is None:
        if isinstance(data, dict):
            labels = list(set(data[label_column]))
        else:  # DataFrame or similar
            labels = list(set(data[label_column].iloc if hasattr(data[label_column], 'iloc') else data[label_column]))
    
    # Format labels as a comma-separated string
    labels_str = ", ".join([f'"{label}"' for label in labels])
    
    # Replace {labels} placeholder in task prompt
    task_prompt = task_prompt.format(labels=labels_str)
    
    # Determine pipeline type
    pipeline_type = get_pipeline_type(model_name)
    print(f"Loading {model_name} ({get_model_info(model_name)}B parameters) with {pipeline_type} pipeline...")
    
    # Initialize model and tokenizer based on pipeline type
    tokenizer = AutoTokenizer.from_pretrained(model_name, trust_remote_code=True)
    
    if pipeline_type == "text2text-generation":
        model = AutoModelForSeq2SeqLM.from_pretrained(model_name, trust_remote_code=True).to(device)
    else:  # text-generation
        model = AutoModelForCausalLM.from_pretrained(model_name, trust_remote_code=True).to(device)
    
    # Create the appropriate pipeline
    classification_pipe = pipeline(
        pipeline_type,
        model=model,
        tokenizer=tokenizer,
        device=0 if device == "cuda" else -1
    )
    
    # Configure generation parameters for deterministic output
    gen_config = {
        "max_new_tokens": max_new_tokens,
        "temperature": 0.0,  # Zero temperature for deterministic output
        "do_sample": False,  # Don't sample
        "num_beams": 1,      # No beam search needed for deterministic output
        "top_p": 1.0,        # No nucleus sampling
        "top_k": 0,          # No top-k filtering
        "repetition_penalty": 1.0,  # No repetition penalty
    }
    
    # Add pipeline-specific config
    if pipeline_type == "text-generation":
        gen_config["return_full_text"] = False
    
    # Get data length depending on the data type
    if hasattr(data, '__len__'):
        data_length = len(data)
    else:
        data_length = len(data[text_column])
    
    # Evaluate
    correct, total = 0, 0
    predictions = []
    
    for i in tqdm(range(0, data_length, stride)):
        # Access text based on data type
        if isinstance(data, dict):
            text = data[text_column][i]
            true_label = data[label_column][i]
        else:  # DataFrame or similar
            text = data[text_column].iloc[i] if hasattr(data[text_column], 'iloc') else data[text_column][i]
            true_label = data[label_column].iloc[i] if hasattr(data[label_column], 'iloc') else data[label_column][i]
        
        # Format input based on pipeline type
        if pipeline_type == "text2text-generation":
            # For encoder-decoder models, combine system and task prompt
            input_text = f"{system_prompt}\n{task_prompt}{text}"
            raw_prediction = classification_pipe(input_text, **gen_config)[0]['generated_text']
        else:
            # For decoder-only models, use message format
            messages = [
                {"role": "system", "content": system_prompt},
                {"role": "user", "content": f"{task_prompt}{text}"},
            ]
            raw_prediction = classification_pipe(messages, **gen_config)[0]['generated_text']
        
        # Clean prediction to match expected format (remove extra spaces, quotes, periods)
        prediction = raw_prediction.strip().strip('"\'.,').lower()
        
        # For exact matching, we also convert true label to lowercase
        true_label_str = str(true_label).lower()
        
        # Store prediction
        predictions.append(prediction)
        
        # Check if prediction matches true label
        if prediction in true_label_str or true_label_str in prediction:
            correct += 1
        total += 1
    
    # Calculate accuracy
    accuracy = correct / total if total > 0 else 0
    
    return {
        "accuracy": accuracy,
        "correct": correct,
        "total": total,
        "model": model_name,
        "params_billions": get_model_info(model_name),
        "pipeline_type": pipeline_type,
        "predictions": predictions
    }


# Example usage:
if __name__ == "__main__":
    # Print available models
    print("Available models (all under 5B parameters):")
    for model, params in sorted(get_model_info().items(), key=lambda x: x[1]):
        print(f"- {model}: {params}B parameters")
    
    # Example for sentiment classification
    # results = eval_llm(
    #     data=validation_data,
    #     system_prompt="You are a classifier that outputs exactly one word from the provided options. No explanations.",
    #     task_prompt="Classify the following text as either \"positive\" or \"negative\". Only respond with one word. Text: ",
    #     model_name="microsoft/Phi-3-mini-4k-instruct",
    #     labels=["positive", "negative"]
    # )
    # print(f"Accuracy: {results['accuracy']:.2%} ({results['correct']}/{results['total']})")



"""
UNet implementation based on "U-Net: Convolutional Networks for Biomedical Image Segmentation"
https://arxiv.org/abs/1505.04597
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Optional, List, Union, Dict


class DoubleConv(nn.Module):
    """Double convolution block: conv -> bn -> relu -> conv -> bn -> relu"""
    
    def __init__(
        self, 
        in_channels: int, 
        out_channels: int, 
        mid_channels: Optional[int] = None,
        kernel_size: int = 3,
        padding: int = 1,
        bias: bool = False
    ) -> None:
        super().__init__()
        if mid_channels is None:
            mid_channels = out_channels
            
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=kernel_size, padding=padding, bias=bias),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=kernel_size, padding=padding, bias=bias),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.double_conv(x)


class Down(nn.Module):
    """Downscaling with maxpool then double conv"""
    
    def __init__(
        self, 
        in_channels: int, 
        out_channels: int, 
        pooling_size: int = 2
    ) -> None:
        super().__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(pooling_size),
            DoubleConv(in_channels, out_channels)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.maxpool_conv(x)


class Up(nn.Module):
    """Upscaling then double conv"""
    
    def __init__(
        self, 
        in_channels: int, 
        out_channels: int, 
        bilinear: bool = True
    ) -> None:
        super().__init__()

        # if bilinear, use the normal convolutions to reduce the number of channels
        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
            self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)
        else:
            self.up = nn.ConvTranspose2d(in_channels, in_channels // 2, kernel_size=2, stride=2)
            self.conv = DoubleConv(in_channels, out_channels)

    def forward(self, x1: torch.Tensor, x2: torch.Tensor) -> torch.Tensor:
        x1 = self.up(x1)
        
        # Adjust dimensions if needed
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2,
                        diffY // 2, diffY - diffY // 2])
        
        # Concatenate along the channel dimension
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    """Final 1x1 convolution layer"""
    
    def __init__(self, in_channels: int, out_channels: int) -> None:
        super().__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.conv(x)


class UNet(nn.Module):
    """UNet architecture for image segmentation tasks"""
    
    def __init__(
        self, 
        in_channels: int = 3, 
        out_channels: int = 1,
        depth: int = 5,
        base_features: int = 64, 
        bilinear: bool = True,
        features: Optional[List[int]] = None
    ) -> None:
        """
        Args:
            in_channels: Number of input channels (e.g., 3 for RGB)
            out_channels: Number of output channels (e.g., 1 for binary segmentation)
            depth: Depth of the UNet (number of down/up operations)
            base_features: Number of features in the first layer
            bilinear: Whether to use bilinear upsampling (True) or transposed convolutions (False)
            features: List of feature dimensions for each layer (if None, generated based on depth and base_features)
        """
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.bilinear = bilinear
        
        # Generate feature dimensions for each level if not provided
        if features is None:
            features = [base_features * (2 ** i) for i in range(depth)]
        
        # Initial double convolution
        self.inc = DoubleConv(in_channels, features[0])
        
        # Downsampling path
        self.downs = nn.ModuleList()
        for i in range(len(features) - 1):
            self.downs.append(Down(features[i], features[i + 1]))
        
        # Upsampling path
        self.ups = nn.ModuleList()
        for i in reversed(range(1, len(features))):
            in_feat = features[i]
            out_feat = features[i - 1]
            self.ups.append(Up(in_feat, out_feat, bilinear))
        
        # Final convolution
        self.outc = OutConv(features[0], out_channels)
        
        # Initialize weights
        self._init_weights()
        
    def _init_weights(self) -> None:
        """Initialize model weights for better convergence"""
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Initial feature extraction
        x1 = self.inc(x)
        
        # Contracting path with skip connections
        skip_connections = [x1]
        for down in self.downs:
            x = down(skip_connections[-1])
            skip_connections.append(x)
        
        # Remove the last feature map (bottom of the U)
        x = skip_connections.pop()
        
        # Expansive path
        for up in self.ups:
            skip = skip_connections.pop()
            x = up(x, skip)
            
        # Final convolution
        return self.outc(x)

def train_unet(
    model: nn.Module,
    train_loader: torch.utils.data.DataLoader,
    val_loader: torch.utils.data.DataLoader,
    criterion: nn.Module,
    optimizer: torch.optim.Optimizer,
    device: torch.device,
    num_epochs: int = 100,
    scheduler: Optional[torch.optim.lr_scheduler._LRScheduler] = None,
    early_stopping_patience: Optional[int] = None,
    save_path: Optional[str] = None,
) -> Dict[str, List[float]]:
    """
    Train the UNet model with validation
    
    Args:
        model: The UNet model to train
        train_loader: DataLoader for training data
        val_loader: DataLoader for validation data
        criterion: Loss function
        optimizer: Optimizer for training
        device: Device to train on (cuda/cpu)
        num_epochs: Number of epochs to train
        scheduler: Learning rate scheduler (optional)
        early_stopping_patience: Number of epochs to wait before early stopping (optional)
        save_path: Path to save the best model (optional)
        
    Returns:
        Dictionary containing training history
    """
    model.to(device)
    history = {
        'train_loss': [],
        'val_loss': [],
        'best_val_loss': float('inf'),
        'best_epoch': 0
    }
    
    no_improve_count = 0
    
    for epoch in range(num_epochs):
        # Training phase
        model.train()
        train_loss = 0.0
        for images, masks in train_loader:
            images, masks = images.to(device), masks.to(device)
            
            # Forward pass
            outputs = model(images)
            loss = criterion(outputs, masks)
            
            # Backward pass and optimize
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            
            train_loss += loss.item() * images.size(0)
        
        train_loss /= len(train_loader.dataset)
        history['train_loss'].append(train_loss)
        
        # Validation phase
        model.eval()
        val_loss = 0.0
        with torch.no_grad():
            for images, masks in val_loader:
                images, masks = images.to(device), masks.to(device)
                outputs = model(images)
                loss = criterion(outputs, masks)
                val_loss += loss.item() * images.size(0)
                
        val_loss /= len(val_loader.dataset)
        history['val_loss'].append(val_loss)
        
        # Print progress
        print(f'Epoch {epoch+1}/{num_epochs} | Train Loss: {train_loss:.4f} | Val Loss: {val_loss:.4f}')
        
        # Learning rate scheduling
        if scheduler is not None:
            scheduler.step(val_loss)
            
        # Save best model
        if val_loss < history['best_val_loss']:
            history['best_val_loss'] = val_loss
            history['best_epoch'] = epoch
            no_improve_count = 0
            
            if save_path:
                torch.save(model.state_dict(), save_path)
                print(f"Model saved to {save_path}")
        else:
            no_improve_count += 1
        
        # Early stopping
        if early_stopping_patience and no_improve_count >= early_stopping_patience:
            print(f"Early stopping at epoch {epoch+1}")
            break
    
    return history


# Example usage
if __name__ == "__main__":
    import torch.optim as optim
    from torch.utils.data import DataLoader, random_split
    from torchvision import transforms
    from torchvision.datasets import ImageFolder
    
    # Define device
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    
    # Create the UNet model
    model = UNet(
        in_channels=3,        # RGB images
        out_channels=1,       # Binary segmentation
        depth=4,              # 4 down/up operations
        base_features=64,     # Start with 64 features
        bilinear=True         # Use bilinear upsampling
    )
    
    # Example dataset and transforms (replace with your actual dataset)
    transform = transforms.Compose([
        transforms.Resize((256, 256)),
        transforms.ToTensor(),
        transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
    ])
    
    # Define loss function and optimizer
    criterion = nn.BCEWithLogitsLoss()
    optimizer = optim.Adam(model.parameters(), lr=1e-4)
    scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer, 'min', patience=5, factor=0.5)
    
    # Example code for dataset preparation (customize for your data)
    """
    # Load and split dataset
    dataset = YourDataset(root_dir='path/to/data', transform=transform)
    train_size = int(0.8 * len(dataset))
    val_size = len(dataset) - train_size
    train_dataset, val_dataset = random_split(dataset, [train_size, val_size])
    
    # Create data loaders
    train_loader = DataLoader(train_dataset, batch_size=8, shuffle=True, num_workers=4)
    val_loader = DataLoader(val_dataset, batch_size=8, shuffle=False, num_workers=4)
    
    # Train the model
    history = train_unet(
        model=model,
        train_loader=train_loader,
        val_loader=val_loader,
        criterion=criterion,
        optimizer=optimizer,
        device=device,
        num_epochs=100,
        scheduler=scheduler,
        early_stopping_patience=10,
        save_path='best_unet_model.pth'
    )
    
    # Load best model for inference
    model.load_state_dict(torch.load('best_unet_model.pth'))
    model.eval()
    
    # Example inference
    with torch.no_grad():
        image = next(iter(val_loader))[0][0:1].to(device)  # Get a single image
        prediction = torch.sigmoid(model(image)) > 0.5     # Apply sigmoid and threshold
        # Save or visualize the prediction
    """



from transformers import AutoImageProcessor, AutoModelForImageClassification
import torch
from torch.utils.data import DataLoader
from tqdm import tqdm
from typing import Dict, List, Optional, Union
import os
from PIL import Image

# Dictionary of common Vision Transformer models with parameter counts (millions)
MODEL_PARAMS = {
    "google/vit-base-patch16-224": 86,
    "google/vit-large-patch16-224": 307,
    "google/vit-huge-patch14-224": 632,
    "facebook/deit-tiny-patch16-224": 5,
    "facebook/deit-small-patch16-224": 22,
    "facebook/deit-base-patch16-224": 86,
    "microsoft/beit-base-patch16-224": 86,
    "microsoft/beit-large-patch16-224": 304,
    "facebook/deit-base-distilled-patch16-224": 87,
    "WinKawaks/vit-small-patch16-224": 22,
}

def load_vit(
    model_name: str = "google/vit-base-patch16-224",
    num_classes: Optional[int] = None,
    device: Optional[str] = None
) -> tuple:
    """
    Load a Vision Transformer model with specified parameters.
    
    Args:
        model_name: HuggingFace model identifier
        num_classes: Number of output classes (if None, uses model's default)
        device: Computing device (auto-detected if None)
        
    Returns:
        Tuple of (model, processor)
    """
    # Determine device
    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    
    # Load processor and model
    processor = AutoImageProcessor.from_pretrained(model_name, use_fast=True)
    model = AutoModelForImageClassification.from_pretrained(
        model_name,
        num_labels=num_classes,
        ignore_mismatched_sizes=True
    ).to(device)
    
    return model, processor

def get_model_info(model_name: Optional[str] = None) -> Union[Dict[str, float], float]:
    """
    Get parameter count information for Vision Transformer models.
    
    Args:
        model_name: Specific model to get information about. If None, returns all models.
        
    Returns:
        Dictionary of all models with parameter counts, or a single count for the specified model.
    """
    if model_name is None:
        return MODEL_PARAMS
    
    if model_name in MODEL_PARAMS:
        return MODEL_PARAMS[model_name]
    
    return f"Model {model_name} not found in the parameter dictionary."

def _validate_model(model, val_loader, device, max_samples=None):
    """
    Helper function for validation during training
    
    Args:
        model: Model to validate
        val_loader: Validation DataLoader
        device: Computing device
        max_samples: Maximum number of samples to validate on (None = all samples)
    """
    model.eval()
    val_loss = 0.0
    correct = 0
    total = 0
    
    sample_count = 0
    with torch.no_grad():
        for batch in val_loader:
            # Get images and labels
            images = batch['pixel_values'].to(device) if 'pixel_values' in batch else batch[0].to(device)
            labels = batch['labels'].to(device) if 'labels' in batch else batch[1].to(device)
            
            # Forward pass
            outputs = model(images, labels=labels)
            val_loss += outputs.loss.item()
            
            # Calculate accuracy if logits are available
            if hasattr(outputs, 'logits'):
                _, predicted = torch.max(outputs.logits, 1)
                batch_size = labels.size(0)
                total += batch_size
                correct += (predicted == labels).sum().item()
            
            # Count samples and break if we've reached max_samples
            sample_count += batch_size
            if max_samples is not None and sample_count >= max_samples:
                break
    
    # Calculate average loss and accuracy
    batch_count = len(val_loader) if max_samples is None else ((sample_count-1) // batch_size + 1)
    avg_loss = val_loss / batch_count if batch_count > 0 else 0
    accuracy = 100 * correct / total if total > 0 else 0
    
    return avg_loss, accuracy

def train_vit(
    model,
    train_loader: DataLoader,
    val_loader: Optional[DataLoader] = None,
    num_epochs: int = 3,
    learning_rate: float = 1e-4,
    patience: int = 5,
    checkpoint_dir: Optional[str] = None,
    device: Optional[str] = None,
    validation_freq: int = 0,  # 0 means evaluate only at end of epochs
    max_val_samples: Optional[int] = None  # Number of samples to use for validation
) -> Dict:
    """
    Train a Vision Transformer model with early stopping and checkpointing.
    
    Args:
        model: The model to train
        train_loader: DataLoader containing training data
        val_loader: DataLoader containing validation data
        num_epochs: Maximum number of training epochs
        learning_rate: Learning rate for optimization
        patience: Number of checks with no improvement after which to stop
        checkpoint_dir: Directory to save model checkpoints
        device: Computing device (auto-detected if None)
        validation_freq: How often to check validation (in steps, 0 = only at epoch end)
        max_val_samples: Maximum samples to use for validation (None = all samples)
        
    Returns:
        Dictionary with training metrics
    """
    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    
    # Setup model and optimizer
    model.to(device)
    model.train()
    optimizer = torch.optim.AdamW(model.parameters(), lr=learning_rate)
    
    # Initialize metrics
    results = {
        "train_loss_history": [],
        "val_loss_history": [] if val_loader else None
    }
    
    # Early stopping and checkpoint variables
    best_val_loss = float('inf')
    no_improve_count = 0
    # Set validation frequency if not specified
    if validation_freq == 0 and val_loader:
        validation_freq = len(train_loader)  # Once per epoch
    elif validation_freq == 0:
        validation_freq = float('inf')  # Never check within epochs
    
    # Create checkpoint directory
    if checkpoint_dir and not os.path.exists(checkpoint_dir):
        os.makedirs(checkpoint_dir)
    
    # Track global step for logging
    global_step = 0
    stop_training = False
    
    for epoch in range(num_epochs):
        if stop_training:
            break
            
        running_loss = 0.0
        correct = 0
        total = 0
        
        progress_bar = tqdm(train_loader, desc=f"Epoch {epoch+1}/{num_epochs}")
        for batch_idx, batch in enumerate(progress_bar):
            # Get images and labels
            images = batch['pixel_values'].to(device) if 'pixel_values' in batch else batch[0].to(device)
            labels = batch['labels'].to(device) if 'labels' in batch else batch[1].to(device)
            
            # Forward and backward pass
            optimizer.zero_grad()
            outputs = model(images, labels=labels)
            loss = outputs.loss
            loss.backward()
            optimizer.step()
            
            # Track metrics
            running_loss += loss.item()
            global_step += 1
            
            # Calculate accuracy if logits are available
            if hasattr(outputs, 'logits'):
                _, predicted = torch.max(outputs.logits, 1)
                batch_size = labels.size(0)
                total += batch_size
                correct += (predicted == labels).sum().item()
                
                # Update progress bar
                progress_bar.set_postfix({
                    'loss': running_loss / (batch_idx + 1),
                    'acc': 100 * correct / total if total > 0 else 0
                })
            
            # Validate within epoch if configured
            if val_loader and global_step % validation_freq == 0:
                val_loss, val_acc = _validate_model(model, val_loader, device, max_val_samples)
                results["val_loss_history"].append(val_loss)
                
                print(f"Step {global_step}: Val Loss: {val_loss:.4f}, Val Acc: {val_acc:.2f}%")
                
                # Check for improvement
                if val_loss < best_val_loss:
                    best_val_loss = val_loss
                    no_improve_count = 0
                    
                    # Save best model checkpoint
                    if checkpoint_dir:
                        checkpoint_path = os.path.join(checkpoint_dir, "best_model.pt")
                        torch.save({
                            'epoch': epoch,
                            'global_step': global_step,
                            'model_state_dict': model.state_dict(),
                            'optimizer_state_dict': optimizer.state_dict(),
                            'val_loss': val_loss,
                        }, checkpoint_path)
                        print(f"Saved new best model with val_loss: {val_loss:.4f}")
                else:
                    no_improve_count += 1
                
                # Early stopping check
                if no_improve_count >= patience:
                    print(f"Early stopping triggered after {global_step} steps")
                    stop_training = True
                    break
                
                # Set model back to training mode
                model.train()
        
        # Record epoch training metrics
        epoch_train_loss = running_loss / len(train_loader)
        epoch_train_acc = 100 * correct / total if total > 0 else 0
        results["train_loss_history"].append(epoch_train_loss)
        
        print(f"Epoch {epoch+1}/{num_epochs}, Train Loss: {epoch_train_loss:.4f}, Train Acc: {epoch_train_acc:.2f}%")
        
        # Full validation at end of epoch if not done in last step
        if val_loader and not stop_training and global_step % validation_freq != 0:
            val_loss, val_acc = _validate_model(model, val_loader, device, max_val_samples)
            if results["val_loss_history"] is None:
                results["val_loss_history"] = []
            results["val_loss_history"].append(val_loss)
            
            print(f"Epoch {epoch+1}: Val Loss: {val_loss:.4f}, Val Acc: {val_acc:.2f}%")
            
            # Check for improvement
            if val_loss < best_val_loss:
                best_val_loss = val_loss
                no_improve_count = 0
                
                # Save best model checkpoint
                if checkpoint_dir:
                    checkpoint_path = os.path.join(checkpoint_dir, "best_model.pt")
                    torch.save({
                        'epoch': epoch,
                        'global_step': global_step,
                        'model_state_dict': model.state_dict(),
                        'optimizer_state_dict': optimizer.state_dict(),
                        'val_loss': val_loss,
                    }, checkpoint_path)
                    print(f"Saved new best model with val_loss: {val_loss:.4f}")
            else:
                no_improve_count += 1
            
            # Early stopping check
            if no_improve_count >= patience:
                print(f"Early stopping triggered after {epoch+1} epochs")
                break
    
    # Load best model if available
    if checkpoint_dir and os.path.exists(os.path.join(checkpoint_dir, "best_model.pt")):
        checkpoint = torch.load(os.path.join(checkpoint_dir, "best_model.pt"))
        model.load_state_dict(checkpoint['model_state_dict'])
        print(f"Loaded best model with validation loss: {checkpoint['val_loss']:.4f}")
    
    return results

def eval_vit(
    model,
    test_loader: DataLoader,
    device: Optional[str] = None,
    max_samples: Optional[int] = None,
    checkpoint_dir: Optional[str] = None
) -> Dict:
    """
    Evaluate a Vision Transformer model on a test dataset.
    
    Args:
        model: The model to evaluate
        test_loader: DataLoader containing test data
        device: Computing device (auto-detected if None)
        max_samples: Maximum number of samples to evaluate on (None = all samples)
        checkpoint_dir: Directory to load model checkpoint from (if None, uses model as-is)
        
    Returns:
        Dictionary with evaluation metrics
    """
    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    
    model.to(device)
    
    # Load checkpoint if provided
    if checkpoint_dir and os.path.exists(os.path.join(checkpoint_dir, "best_model.pt")):
        checkpoint = torch.load(os.path.join(checkpoint_dir, "best_model.pt"), map_location=device)
        model.load_state_dict(checkpoint['model_state_dict'])
        print(f"Loaded best model from {checkpoint_dir} with validation loss: {checkpoint['val_loss']:.4f}")
    
    model.eval()
    
    results = {
        "correct": 0,
        "total": 0,
        "loss": 0.0,
        "class_accuracy": {}
    }
    
    sample_count = 0
    with torch.no_grad():
        for batch in tqdm(test_loader, desc="Evaluating"):
            # Get images and labels
            images = batch['pixel_values'].to(device) if 'pixel_values' in batch else batch[0].to(device)
            labels = batch['labels'].to(device) if 'labels' in batch else batch[1].to(device)
            
            # Forward pass
            outputs = model(images, labels=labels)
            results["loss"] += outputs.loss.item()
            
            # Get predictions
            _, predicted = torch.max(outputs.logits, 1)
            
            # Update metrics
            batch_size = labels.size(0)
            results["total"] += batch_size
            correct = (predicted == labels).sum().item()
            results["correct"] += correct
            
            # Track per-class accuracy
            for i in range(batch_size):
                true_label = labels[i].item()
                pred_label = predicted[i].item()
                
                label_str = str(true_label)
                if label_str not in results["class_accuracy"]:
                    results["class_accuracy"][label_str] = {"correct": 0, "total": 0}
                
                results["class_accuracy"][label_str]["total"] += 1
                if true_label == pred_label:
                    results["class_accuracy"][label_str]["correct"] += 1
            
            # Count samples and break if we've reached max_samples
            sample_count += batch_size
            if max_samples is not None and sample_count >= max_samples:
                break
    
    # Calculate overall metrics
    batch_count = len(test_loader) if max_samples is None else ((sample_count-1) // batch_size + 1)
    results["accuracy"] = results["correct"] / results["total"] if results["total"] > 0 else 0
    results["avg_loss"] = results["loss"] / batch_count if batch_count > 0 else 0
    
    # Calculate per-class accuracy
    for cls in results["class_accuracy"]:
        cls_metrics = results["class_accuracy"][cls]
        if cls_metrics["total"] > 0:
            cls_metrics["accuracy"] = cls_metrics["correct"] / cls_metrics["total"]
        else:
            cls_metrics["accuracy"] = 0.0
    
    return results

# Example usage with dataloaders:
if __name__ == "__main__":
    num_classes = 10  # Example: number of classes in your dataset
    
    # Load model and processor
    model, processor = load_vit(
        model_name="google/vit-base-patch16-224",
        num_classes=num_classes
    )
    
    # Example: Create train, validation and test loaders (you would need to implement these)
    # train_loader = create_dataloader(...)
    # val_loader = create_dataloader(...)
    # test_loader = create_dataloader(...)
    
    # Train model with early stopping and checkpointing
    # train_results = train_vit(
    #     model, 
    #     train_loader,
    #     val_loader=val_loader,
    #     num_epochs=10,
    #     patience=5,
    #     checkpoint_dir="./checkpoints",
    #     validation_freq=5,  # Check validation every 100 steps,
    #     max_val_samples=1000  # Use only 1000 samples for validation
    # )
    
    # Evaluate model
    # eval_results = eval_vit(model, test_loader, checkpoint_dir="./checkpoints")
    # print(f"Test Accuracy: {eval_results['accuracy']:.2%} ({eval_results['correct']}/{eval_results['total']})")
